"""Code to use AO calibrate s"""

from __future__ import annotations  # used to keep mypy/pylance happy in AOSolutions

import struct
from argparse import ArgumentParser
from pathlib import Path
from typing import (
    Any,
    Collection,
    Iterable,
    NamedTuple,
)

import matplotlib.pyplot as plt
import numpy as np

from flint.bptools.preflagger import (
    construct_jones_over_max_amp_flags,
    construct_mesh_ant_flags,
    flag_mean_residual_amplitude,
    flag_mean_xxyy_amplitude_ratio,
    flag_outlier_phase,
    flags_over_threshold,
)
from flint.bptools.smoother import (
    divide_bandpass_by_ref_ant_preserve_phase,
    smooth_bandpass_complex_gains,
)
from flint.exceptions import PhaseOutlierFitError
from flint.logging import logger
from flint.ms import consistent_ms, get_beam_from_ms
from flint.naming import get_aocalibrate_output_path
from flint.options import MS, BaseOptions
from flint.sclient import run_singularity_command
from flint.utils import create_directory


class CalibrateOptions(BaseOptions):
    """Structure used to represent options into the `calibrate` program

    These attributes have the same names as options into the `calibrate`
    command.
    """

    datacolumn: str
    """The name of the datacolumn that will be calibrates"""
    m: Path
    """The path to the model file used to calibtate"""
    minuv: float | None = None
    """The minimum distance in meters that is"""
    maxuv: float | None = None
    """The maximum distance in meters that is"""
    i: int | None = 100
    """The number of iterations that may be performed"""
    p: tuple[Path, Path] | None = None
    """Plot output names for the amplitude gain and phases"""


class CalibrateCommand(BaseOptions):
    """The AO Calibrate command and output path of the corresponding solutions file"""

    cmd: str
    """The calibrate command that will be executed
    """
    solution_path: Path
    """The output path of the solutions file
    """
    ms: MS
    """The measurement set to have solutions derived for"""
    model: Path
    """Path to the model that would be used to calibrate against"""
    preflagged: bool = False
    """Indicates whether the solution file has gone through preflagging routines. """


class ApplySolutions(BaseOptions):
    """The applysolutions command to execute"""

    cmd: str
    """The command that will be executed"""
    solution_path: Path
    """Location of the solutions file to apply"""
    ms: MS
    """The measurement set that will have the solutions applied to"""


# TODO: Rename the bandpass attribute?
class AOSolutions(NamedTuple):
    """Structure to load an AO-style solutions file"""

    path: Path
    """Path of the solutions file loaded"""
    nsol: int
    """Number of time solutions"""
    nant: int
    """Number of antenna in the solution file"""
    nchan: int
    """Number of channels in the solution file"""
    npol: int
    """Number of polarisations in the file"""
    bandpass: np.ndarray
    """Complex data representing the antenna Jones. Shape is (nsol, nant, nchan, npol)"""

    # TODO: Need tocorporate the start and end times into this header

    @classmethod
    def load(cls, path: Path) -> AOSolutions:
        """Load in an AO-stule solution file. See `load_solutions_file`, which is
        internally used.
        """
        return load_aosolutions_file(solutions_path=path)

    def save(self, output_path: Path) -> Path:
        """Save the instance of AOSolution to a standard aosolution binary file

        Args:
            output_path (Path): Location to write the file to

        Returns:
            Path: Location the file was written to
        """
        return save_aosolutions_file(aosolutions=self, output_path=output_path)

    def plot_solutions(self, ref_ant: int | None = 0) -> Iterable[Path]:
        """Plot the solutions of all antenna for the first time-interval
        in the aosolutions file. The XX and the YY will be plotted.

        Args:
            ref_ant (Optional[int], optional): Reference antenna to use. If None is specified there is no division by a reference antenna.  Defaults to 0.

        Returns:
            Iterable[Path]: Path to the phase and amplited plots created.
        """
        # TODO: Change call signature to pass straight through
        return plot_solutions(solutions=self, ref_ant=ref_ant)


def fill_between_flags(
    ax: plt.Axes,
    flags: np.ndarray,
    values: np.ndarray | None = None,
    direction: str = "x",
) -> None:
    """Plot vertical or horizontal lines where data are flagged.

    NOTE: This is pretty inefficient and not intended for regular use.

    Args:
        ax (plt.Axes): Axes object to plot lines on
        flags (np.ndarray): Flags to consider. If `True`, plot.
        values (Optional[np.ndarray], optional): The values to plot at. Useful if the position does not map to location. Defaults to None.
        direction (str, optional): If `x` use axvline, if `y` use axhline. Defaults to "x".
    """
    values = values if values else np.arange(len(flags))

    mask = np.argwhere(flags)
    plot_vals = values[mask]
    func = ax.axvline if direction == "x" else ax.axhline

    for v in plot_vals:
        func(v, color="black", alpha=0.3)


def plot_solutions(
    solutions: Path | AOSolutions, ref_ant: int | None = 0
) -> Collection[Path]:
    """Plot solutions for AO-style solutions

    Args:
        solutions (Path): Path to the solutions file
        ref_ant (Optional[int], optional): Reference antenna to use. If None is specified there is no division by a reference antenna.  Defaults to 0.

    Return:
        Collection[Path] -- The paths of the two plots createda
    """
    ao_sols = (
        AOSolutions.load(path=solutions) if isinstance(solutions, Path) else solutions
    )
    solutions_path = ao_sols.path
    logger.info(f"Plotting {solutions_path}")

    if ao_sols.nsol > 1:
        logger.warning(f"Found {ao_sols.nsol} intervals, plotting the first. ")
    plot_sol = 0  # The first time interval

    data = ao_sols.bandpass[plot_sol]
    if ref_ant is not None and ref_ant < 0:
        ref_ant = select_refant(bandpass=ao_sols.bandpass)
        logger.info(f"Overwriting reference antenna selection, using {ref_ant=}")

    if ref_ant is not None:
        data = divide_bandpass_by_ref_ant_preserve_phase(
            complex_gains=ao_sols.bandpass[plot_sol], ref_ant=ref_ant
        )

    amplitudes = np.abs(data)
    phases = np.angle(data, deg=True)
    channels = np.arange(ao_sols.nchan)

    ncolumns = 6
    nrows = ao_sols.nant // ncolumns
    if ncolumns * nrows < ao_sols.nant:
        nrows += 1
    logger.debug(f"Plotting {plot_sol=} with {ncolumns=} {nrows=}")

    fig_amp, axes_amp = plt.subplots(nrows, ncolumns, figsize=(15, 9))
    fig_ratio, axes_ratio = plt.subplots(nrows, ncolumns, figsize=(15, 9))
    fig_phase, axes_phase = plt.subplots(nrows, ncolumns, figsize=(15, 9))

    for y in range(nrows):
        for x in range(ncolumns):
            ant = y * nrows + x

            amps_xx = amplitudes[ant, :, 0]
            amps_yy = amplitudes[ant, :, 3]
            phase_xx = phases[ant, :, 0]
            phase_yy = phases[ant, :, 3]

            ratio = amps_xx / amps_yy

            if any([np.sum(~np.isfinite(amps)) == 0 for amps in (amps_xx, amps_yy)]):
                logger.warning(f"No valid data for {ant=}")
                continue

            max_amp_xx = (
                np.nanmax(amps_xx[np.isfinite(amps_xx)])
                if any(np.isfinite(amps_xx))
                else -1
            )
            max_amp_yy = (
                np.nanmax(amps_yy[np.isfinite(amps_yy)])
                if any(np.isfinite(amps_yy))
                else -1
            )
            min_amp_xx = (
                np.nanmin(amps_xx[np.isfinite(amps_xx)])
                if any(np.isfinite(amps_xx))
                else -1
            )
            min_amp_yy = (
                np.nanmin(amps_yy[np.isfinite(amps_yy)])
                if any(np.isfinite(amps_yy))
                else -1
            )
            ax_a, ax_p = axes_amp[y, x], axes_phase[y, x]
            ax_a = axes_amp[y, x]
            ax_r = axes_ratio[y, x]
            ax_a.plot(
                channels,
                amps_xx,
                marker=None,
                color="tab:blue",
                label="X" if y == 0 and x == 0 else None,
            )
            ax_a.plot(
                channels,
                amps_yy,
                marker=None,
                color="tab:red",
                label="Y" if y == 0 and x == 0 else None,
            )
            ax_r.plot(
                channels,
                ratio,
                marker=None,
                color="tab:green",
                label="X/Y" if y == 0 and x == 0 else None,
            )

            ax_a.set(
                ylim=(
                    min(min_amp_xx, min_amp_yy) * 0.9,
                    max(max_amp_xx, max_amp_yy) * 1.1,
                )
            )
            ax_a.axhline(1, color="black", linestyle="--", linewidth=0.5)
            ax_a.set_title(f"ant{ant:02d}", fontsize=8)
            fill_between_flags(ax_a, ~np.isfinite(amps_yy) | ~np.isfinite(amps_xx))

            ax_r.set(ylim=(0.8, 1.2))
            ax_r.axhline(1, color="black", linestyle="--", linewidth=0.5)
            ax_r.set_title(f"ant{ant:02d}", fontsize=8)
            fill_between_flags(ax_r, ~np.isfinite(amps_yy) | ~np.isfinite(amps_xx))

            ax_p.plot(
                channels,
                phase_xx,
                marker=None,
                color="tab:blue",
                label="X" if y == 0 and x == 0 else None,
            )
            ax_p.plot(
                channels,
                phase_yy,
                marker=None,
                color="tab:red",
                label="Y" if y == 0 and x == 0 else None,
            )
            ax_p.set(ylim=(-200, 200))
            ax_p.set_title(f"ak{ant:02d}", fontsize=8)
            fill_between_flags(ax_p, ~np.isfinite(phase_yy) | ~np.isfinite(phase_xx))

    fig_amp.legend()
    fig_phase.legend()
    fig_ratio.legend()

    fig_amp.suptitle(f"{ao_sols.path.name} - Amplitudes")
    fig_phase.suptitle(f"{ao_sols.path.name} - Phases")
    fig_ratio.suptitle(f"{ao_sols.path.name} - Amplitude Ratios")

    fig_amp.tight_layout()
    fig_ratio.tight_layout()
    fig_phase.tight_layout()

    out_amp = f"{solutions_path.with_suffix('.amplitude.png')!s}"
    logger.info(f"Saving {out_amp}.")
    fig_amp.savefig(out_amp)

    out_phase = f"{solutions_path.with_suffix('.phase.png')!s}"
    logger.info(f"Saving {out_phase}.")
    fig_phase.savefig(out_phase)

    out_ratio = f"{solutions_path.with_suffix('.ratio.png')!s}"
    logger.info(f"Saving {out_ratio}.")
    fig_ratio.savefig(out_ratio)

    return [Path(out_amp), Path(out_phase), Path(out_ratio)]


def save_aosolutions_file(aosolutions: AOSolutions, output_path: Path) -> Path:
    """Save a AOSolutions file to the ao-standard binary format.

    Args:
        aosolutions (ApplySolutions): Instance of the solutions to save
        output_path (Path): Output path to write the files to

    Returns:
        Path: Path the file was written to
    """

    header_format = "8s6I2d"
    header_intro = b"MWAOCAL\0"

    output_dir = output_path.parent
    if not output_dir.exists():
        logger.info(f"Creating {output_dir}.")
        output_dir.mkdir(parents=True)

    logger.info(f"Writing aosolutions to {output_path!s}.")
    with open(str(output_path), "wb") as out_file:
        out_file.write(
            struct.pack(
                header_format,
                header_intro,
                0,  # File type, only 0 mode available
                0,  # Structure type, 0 model available only
                aosolutions.nsol,
                aosolutions.nant,
                aosolutions.nchan,
                aosolutions.npol,
                0.0,  # time start, I don't believe these are used in most use cases
                0.0,  # time end, I don't believe these are used in most use cases
            )
        )
        aosolutions.bandpass.tofile(out_file)

    return output_path


def load_aosolutions_file(solutions_path: Path) -> AOSolutions:
    """Load in an AO-style solutions file

    Args:
        solutions_path (Path): The path of the solutions file to load

    Returns:
        AOSolutions: Structure container the deserialized solutions file
    """

    assert solutions_path.exists() and solutions_path.is_file(), (
        f"{solutions_path!s} either does not exist or is not a file. "
    )
    logger.info(f"Loading {solutions_path}")

    with open(solutions_path) as in_file:
        _junk = np.fromfile(in_file, dtype="<i4", count=2)

        header = np.fromfile(in_file, dtype="<i4", count=10)
        logger.info(f"Header extracted: {header=}")
        file_type = header[0]
        assert file_type == 0, f"Expected file_type of 0, found {file_type}"

        structure_type = header[1]
        assert file_type == 0, f"Expected structure_type of 0, found {structure_type}"

        nsol, nant, nchan, npol = header[2:6]
        sol_shape = (nsol, nant, nchan, npol)

        bandpass = np.fromfile(in_file, dtype="<c16", count=np.prod(sol_shape)).reshape(
            sol_shape
        )
        logger.info(f"Loaded solutions of shape {bandpass.shape}")

        return AOSolutions(
            path=solutions_path,
            nsol=nsol,
            nant=nant,
            nchan=nchan,
            npol=npol,
            bandpass=bandpass,
        )


def find_existing_solutions(
    bandpass_directory: Path,
    use_preflagged: bool = True,
    use_smoothed: bool = False,
) -> list[CalibrateCommand]:
    """Given a directory that contains a collection of bandpass measurement
    sets, attempt to identify a corresponding set of calibrate binary solution
    file.

    This search only supports the use of the default or known preflagger suffix.
    Limited support is provided to specify the expected calibrate suffix.

    These bandpass measurement sets should be processed already - which means just
    the B1936-638 field has been split out of the larger raw MS, flagged and
    calibrated. These steps are expected in order to get to the calibrate
    stage.

    Args:
        bandpass_directory (Path): Directory to search for split bandpass measurement sets
        use_preflagged (bool, optional): Add the pre-flag suffix when searching for solution files. This uses the expected suffix for preflagged solutions. Defaults to True.
        use_smoothed (bool, optional): Add the smoothed bandpass suffix when searching for solution files. This uses the expected suffix for smoothed solutions. Defaults to False.

    Returns:
        List[CalibrateCommand]: Collection of the calibrate command structures that are intended to be used to map the bandpass measurement sets to solution files.
    """
    logger.info(
        f"Searching {bandpass_directory} for existing measurement sets and solutions. "
    )

    bandpass_mss = list(bandpass_directory.glob("*ms"))
    logger.info(f"Found {len(bandpass_mss)} bandpass measurement sets")

    solution_paths = [
        get_aocalibrate_output_path(
            ms_path=bandpass_ms,
            include_preflagger=use_preflagged,
            include_smoother=use_smoothed,
        )
        for bandpass_ms in bandpass_mss
    ]

    # If not all the treasure could be found. At the moment this function will only
    # work if the bandpass solutions were made using the default values.
    assert all([solution_path.exists() for solution_path in solution_paths]), (
        f"Missing solution file constructed from scanning {bandpass_directory}. Check the directory. "
    )

    calibrate_cmds = [
        CalibrateCommand(
            cmd="None",
            ms=MS(ms),
            solution_path=solution_path,
            model=Path("None"),
            preflagged=True,
        )
        for (ms, solution_path) in zip(bandpass_mss, solution_paths)
    ]

    logger.info(f"Constructed {len(calibrate_cmds)} calibrate commands. ")

    return calibrate_cmds


def select_aosolution_for_ms(
    calibrate_cmds: list[CalibrateCommand], ms: MS | Path
) -> Path:
    """Attempt to select an AO-style solution file for a measurement
    set. This can be expanded to include a number of criteria, but
    at present it only searches for a matching beam number between
    the input set of CalibrationCommands and the input MS.

    Args:
        calibrate_cmds (List[CalibrateCommand]): Set of calibration commands, which includes the solution file path and the corresponding MS, as attributes.
        ms (Union[MS, Path]): The measurement sett that needs a solutions file.

    Raises:
        ValueError: Raised when not matching AO-solution file found.

    Returns:
        Path: Path to solution file to apply.
    """
    ms = MS.cast(ms)
    ms_beam = ms.beam if ms.beam is not None else get_beam_from_ms(ms=ms)

    logger.info(f"Will select a solution for {ms.path!s}, {ms_beam=}.")
    logger.info(f"{len(calibrate_cmds)} potential solutions to consider. ")

    for calibrate_cmd in calibrate_cmds:
        logger.info(f"Considering {calibrate_cmd.solution_path!s}.")
        if consistent_ms(ms1=ms, ms2=calibrate_cmd.ms):
            sol_file = calibrate_cmd.solution_path
            break
    else:
        raise ValueError(
            f"No solution file found for {ms.path!s} from {[c.ms.path for c in calibrate_cmds]} found. "
        )

    logger.info(f"Have selected {sol_file!s} for {ms.path!s}. ")
    return sol_file


def calibrate_options_to_command(
    calibrate_options: CalibrateOptions, ms_path: Path, solutions_path: Path
) -> str:
    """Generate a `calibrate` command given an input option set

    Args:
        calibrate_options (CalibrateOptions): The set of `calibrate` options to use
        ms (Path): Path to the measurement set that will be calibrated
        solutions_path (Path): Output path of the solutions file

    Returns:
        str: The command string to execute
    """
    cmd = "calibrate "

    unknowns: list[tuple[Any, Any]] = []

    for key, value in calibrate_options._asdict().items():
        if value is None:
            continue
        elif isinstance(value, (str, Path, int, float)):
            cmd += f"-{key} {value!s} "
        elif isinstance(value, (tuple, list)):
            values = " ".join([str(v) for v in value])
            cmd += f"-{key} {values} "
        else:
            unknowns.append((key, value))

    assert len(unknowns) == 0, (
        f"Unknown types when generating calibrate command: {unknowns}"
    )

    cmd += f"{ms_path!s} {solutions_path!s}"

    return cmd


def create_calibrate_cmd(
    ms: Path | MS,
    calibrate_model: Path,
    solution_path: Path | None = None,
    container: Path | None = None,
    update_calibrate_options: dict[str, Any] | None = None,
    calibrate_data_column: str | None = None,
) -> CalibrateCommand:
    """Generate a typical ao calibrate command. Any extra keyword arguments
    are passed through as additional options to the `calibrate` program.

    Args:
        ms (Union[Path,MS]): The measurement set to calibrate. There needs to be a nominated data_column.
        calibrate_model (Path): Path to a generated calibrate sky-model
        solution_path (Path, optional): The output path of the calibrate solutions file. If None, a default suffix of "calibrate.bin" is used. Defaults to None.
        container (Optional[Path], optional): If a path to a container is supplied the calibrate command is executed immediately. Defaults to None.
        update_calibrate_options (Optional[Dict[str, Any]], optional): Additional options to update the generated CalibrateOptions with. Keys should be attributes of CalibrationOptions. Defaults to None.
        calibrate_data_column(Optional[str], optional): The name of the column to calibrate, overwriting the nominated column set in the MS. If None, the MS.column attribute is used. Defaults to None.

    Raises:
        FileNotFoundError: Raised when calibrate_model can not be found.

    Returns:
        CalibrateCommand: The calibrate command to execute and output solution file
    """
    ms = MS.cast(ms)

    column = ms.column
    if calibrate_data_column:
        logger.info(
            f"Overwriting column to calibrate from {ms.column=} to {calibrate_data_column=}"
        )
        column = calibrate_data_column

    assert column is not None, f"{ms} does not have a nominated data_column"

    logger.info(f"Creating calibrate command for {ms.path}")
    logger.info(f"Will calibrate data column {column}")

    # This is a typical calibrate command.
    # calibrate -minuv 100 -i 50 -datacolumn DATA
    #        -m 2022-04-14_100122_0.calibrate.txt
    #        2022-04-14_100122_0.ms 2022-04-14_100122_0.aocalibrate.bin

    if not calibrate_model.exists():
        raise FileNotFoundError(f"Calibrate model {calibrate_model} not found. ")

    if solution_path is None:
        solution_path = get_aocalibrate_output_path(
            ms_path=ms.path, include_preflagger=False, include_smoother=False
        )

    calibrate_options = CalibrateOptions(
        datacolumn=column, m=calibrate_model, minuv=600
    )
    if update_calibrate_options:
        calibrate_options = calibrate_options.with_options(**update_calibrate_options)

    cmd = calibrate_options_to_command(
        calibrate_options=calibrate_options,
        ms_path=ms.path,
        solutions_path=solution_path,
    )
    logger.debug(f"Constructed calibrate command is {cmd=}")

    calibrate_cmd = CalibrateCommand(
        cmd=cmd, solution_path=solution_path, ms=ms, model=calibrate_model
    )

    if container is not None:
        run_calibrate(calibrate_cmd=calibrate_cmd, container=container)

    return calibrate_cmd


def create_apply_solutions_cmd(
    ms: MS,
    solutions_file: Path,
    output_column: str | None = None,
    container: Path | None = None,
) -> ApplySolutions:
    """Construct the command to apply calibration solutions to a MS
    using an AO calibrate style solutions file.

    The `applysolutions` program does not appear to have the ability to set
    a desured output column name. If the `output_column` specified matches
    the nominated column in `ms`, then `applysolutions` will simply overwrite
    the column with updated data. Otherwise, a `CORRECTED_DATA` column is produced.

    NOTE: Care to be taken when the nominated column is `CORRECTED_DATA`.

    Args:
        ms (MS): Measurement set to have solutions applied to
        solutions_file (Path): Path to the solutions file to apply
        output_column (Optional[str], optional): The desired output column name. See notes above. Defaults to None.
        container (Optional[Path], optional): If a path to a container is supplied the calibrate command is executed immediately. Defaults to None.

    Returns:
        ApplySolutions: Description of applysolutions command, solutions file path and updated MS
    """
    # extract the ms property, if required
    ms = MS.cast(ms)

    assert ms.path.exists(), f"The measurement set {ms} was not found. "
    assert ms.column is not None, f"{ms} does not have a nominated data_column. "
    assert solutions_file.exists(), (
        f"The solutions file {solutions_file} does not exists. "
    )

    input_column = ms.column
    copy_mode = "-nocopy" if input_column == output_column else "-copy"

    logger.info(f"Setting {copy_mode=}.")

    if copy_mode == "-copy":
        output_column = "CORRECTED_DATA"

    cmd = (
        f"applysolutions "
        f"-datacolumn {input_column} "
        f"{copy_mode} "
        f"{ms.path!s} "
        f"{solutions_file!s} "
    )

    logger.info(f"Constructed {cmd=}")

    apply_solutions = ApplySolutions(
        cmd=cmd, solution_path=solutions_file, ms=ms.with_options(column=output_column)
    )

    if container is not None:
        run_apply_solutions(apply_solutions_cmd=apply_solutions, container=container)

    # TODO: If outputcolumn is not CORRECTED_DATA then it should be renamed
    # applysolutions always calls it CORRECTED_DATA

    return apply_solutions


def run_calibrate(calibrate_cmd: CalibrateCommand, container: Path) -> None:
    """Execute a calibrate command within a singularity container

    Args:
        calibrate_cmd (CalibrateCommand): The constructed calibrate command
        container (Path): Location of the container
    """

    assert container.exists(), f"The calibrate container {container} does not exist. "
    assert calibrate_cmd.ms is not None, (
        "When calibrating the 'ms' field attribute must be defined. "
    )

    run_singularity_command(
        image=container,
        command=calibrate_cmd.cmd,
        bind_dirs=[
            calibrate_cmd.solution_path.parent,
            calibrate_cmd.ms.path.parent,
            calibrate_cmd.model.parent,
        ],
    )


def run_apply_solutions(apply_solutions_cmd: ApplySolutions, container: Path) -> None:
    """Will execute the applysolutions command inside the specified singularity
    container.

    Args:
        apply_solutions_cmd (ApplySolutions): The constructed applysolutions command
        container (Path): Location of the existing solutions file
    """

    assert container.exists(), (
        f"The applysolutions container {container} does not exist. "
    )
    assert apply_solutions_cmd.ms.path.exists(), (
        f"The measurement set {apply_solutions_cmd.ms} was not found. "
    )

    run_singularity_command(
        image=container,
        command=apply_solutions_cmd.cmd,
        bind_dirs=[
            apply_solutions_cmd.solution_path.parent.absolute(),
            apply_solutions_cmd.ms.path.parent.absolute(),
        ],
    )


def calibrate_apply_ms(
    ms_path: Path, model_path: Path, container: Path, data_column: str = "DATA"
) -> ApplySolutions:
    """Will create and run a calibration command using AO calibrator, and then apply these solutions.

    Args:
        ms_path (Path): The measurement set that will be calibrated
        model_path (Path): The model file containing sources to calibrate against
        container (Path): Container that has the AO calibtate and applysolutions file.
        data_column (str, optional): The name of the column containing the data to calibrate. Defaults to "DATA".

    Returns:
        Applysolutions: The command, solution binary path and new measurement set structure
    """
    ms = MS(path=ms_path, column=data_column)

    logger.info(f"Will be attempting to calibrate {ms}")

    calibrate_cmd = create_calibrate_cmd(ms=ms, calibrate_model=model_path)

    run_calibrate(calibrate_cmd=calibrate_cmd, container=container.absolute())

    flagged_solutions = flag_aosolutions(
        solutions_path=calibrate_cmd.solution_path,
        ref_ant=0,
        plot_dir=Path(ms_path.parent) / Path("preflagger"),
    )

    apply_solutions_cmd = create_apply_solutions_cmd(
        ms=ms, solutions_file=flagged_solutions.path
    )

    run_apply_solutions(
        apply_solutions_cmd=apply_solutions_cmd, container=container.absolute()
    )

    return apply_solutions_cmd


def apply_solutions_to_ms(
    ms: Path | MS,
    solutions_path: Path,
    container: Path,
    data_column: str = "DATA",
) -> ApplySolutions:
    ms = ms if isinstance(ms, MS) else MS(path=ms, column=data_column)
    logger.info(f"Will attempt to apply {solutions_path!s} to {ms.path!s}.")

    apply_solutions_cmd = create_apply_solutions_cmd(
        ms=ms, solutions_file=solutions_path
    )

    run_apply_solutions(
        apply_solutions_cmd=apply_solutions_cmd, container=container.absolute()
    )

    return apply_solutions_cmd


def select_refant(bandpass: np.ndarray) -> int:
    """Attempt to select an optimal reference antenna. This works in
    a fairly simple way, and simply selects the antenna which is select
    based purely on the number of valid/unflagged solutions in the
    bandpass aosolutions file.

    Args:
        bandpass (np.ndarray): The aosolutions file that has been
        solved for

    Returns:
        int: The index of the reference antenna that should be used.
    """

    assert len(bandpass.shape) == 4, (
        f"Expected a bandpass of shape (times, ant, channels, pol), received {bandpass.shape=}"
    )

    # create the mask of valid solutions
    mask = np.isfinite(bandpass)
    # Sum_mask will be a shape of length 2 (time, ants)
    sum_mask = np.sum(mask, axis=(2, 3))

    # The refant will be the one with the highest number
    max_ant = np.argmax(sum_mask, keepdims=True)

    return max_ant[0][0]


class FlaggedAOSolution(NamedTuple):
    """Hold the final set of flagged solutions and generated plots"""

    path: Path
    """Path to the final set of flagged solutions"""
    plots: Collection[Path]
    """Contains paths to the plots generated throughout the flagging and smoothing procedure"""
    bandpass: np.ndarray
    """The bandpass solutions after flagging, as saved in the solutions file"""


# TODO: These options are too much and should be placed
# into a BaseOptions


def flag_aosolutions(
    solutions_path: Path,
    ref_ant: int = -1,
    flag_cut: float = 3,
    plot_dir: Path | None = None,
    out_solutions_path: Path | None = None,
    smooth_solutions: bool = False,
    plot_solutions_throughout: bool = True,
    smooth_window_size: int = 16,
    smooth_polynomial_order: int = 4,
    mean_ant_tolerance: float = 0.2,
    mesh_ant_flags: bool = False,
    max_gain_amplitude: float | None = None,
) -> FlaggedAOSolution:
    """Will open a previously solved ao-calibrate solutions file and flag additional channels and antennae.

    There are a number of distinct operations applied to the data, which are
    presented in order they are applied.

    If `mesh_ant_flags` is `True`, channels flagged from on channel on a single
    antenna will be applied to all (unless an antenna is completely flagged).
    This happens before any other operation,.

    If `max_gain_amplitude` is not `None` than any Jones with an element
    whose amplitude is above the set value will be flagged.

    Next, an attempt is made to search for channels where the the phase of the
    gain solution are outliers. The phase over frequency is first unwrapped (delay solved for) before the flagging
    statistics are computed.

    If an antenna is over 80% flagged then it is completely removed.

    A low order polynomial (typically order 5) is fit to the amplitudes of the
    Gx and Gy, and if the residuals are sufficiently high then the antenna will
    be flagged.

    If the mean ratio of the Gx and Gy amplitudes for an antenna are higher
    then `mean_ant_tolerance` then the antenna will be flagged.

    Keywords that with the `smooth` prefix are passed to the `smooth_bandpass_complex_gains` function.

    Args:
        solutions_path (Path): Location of the solutions file to examine and flag.
        ref_ant (int, optional): Reference antenna to use, which is important when searching for phase-outliers and to smooth the bandpass. If ref_ant < 0, then an optimal one is selected. Defaults to -1.
        flag_cut (float, optional): Significance of a phase-outlier from the mean (or median) before it should be flagged. Defaults to 3.
        plot_dir (Optional[Path], optional): Where diagnostic flagging plots should be written. If None, no plots will be produced. Defaults to None.
        out_solutions_path (Optional[Path], optional): The output path of the flagged solutions file. If None, the solutions_path provided is used. Defaults to None.
        smooth_solutions (blool, optional): Smooth the complex gain solutions after flaggined. Defaults to False.
        plot_solutions_throughout (bool, Optional): If True, the solutions will be plotted at different stages of processing. Defaults to True.
        smooth_window_size (int, optional): The size of the window function of the savgol filter. Passed directly to savgol. Defaults to 16.
        smooth_polynomial_order (int, optional): The order of the polynomial of the savgol filter. Passed directly to savgol. Defaults to 4.
        mean_ant_tolerance (float, optional): Tolerance of the mean x/y antenna gain ratio test before the antenna is flagged. Defaults to 0.2.
        mesh_ant_flags (bool, optional): If True, a channel is flagged across all antenna if it is flagged for any antenna. Performed before other flagging operations. Defaults to False.
        max_gain_amplitude (Optional[float], optional): If not None, flag the Jones if an antenna has a amplitude gain above this value. Defaults to 10.

    Returns:
        FlaggedAOSolution: Path to the updated solutions file, intermediate solution files and plots along the way
    """
    # TODO: This should be broken down into separate stages. Way too large of a function.
    # TODO: This pirate needs to cull some of this logic out, likely not needed
    # and dead

    solutions = AOSolutions.load(path=solutions_path)
    title = solutions_path.name

    pols = {0: "XX", 1: "XY", 2: "YX", 3: "YY"}

    if plot_dir:
        create_directory(directory=plot_dir)

    # Note that although the solutions variable (an instance of AOSolutions) is immutable,
    # which includes the reference to the numpy array, the _actual_ numpy array is! So,
    # copying the bandpass below is as we are updating the actual array, which will be
    # written back as a new file later.
    bandpass = solutions.bandpass
    logger.info(f"Loaded bandpass, shape is {bandpass.shape}")

    if ref_ant < 0:
        ref_ant = select_refant(bandpass=solutions.bandpass)
        logger.info(f"Overwriting reference antenna selection, using {ref_ant=}")

    plots: list[Path] = []

    if plot_solutions_throughout:
        output_plots = plot_solutions(solutions=solutions_path, ref_ant=ref_ant)
        plots.extend(output_plots)

    if mesh_ant_flags:
        logger.info("Combining antenna flags")
        mask = np.zeros_like(bandpass, dtype=bool)

        for time in range(solutions.nsol):
            mask[time] = construct_mesh_ant_flags(mask=~np.isfinite(bandpass[time]))

        bandpass[mask] = np.nan

    if max_gain_amplitude:
        mask = construct_jones_over_max_amp_flags(
            complex_gains=bandpass, max_amplitude=max_gain_amplitude
        )
        bandpass[mask] = np.nan

    for time in range(solutions.nsol):
        ref_bandpass = divide_bandpass_by_ref_ant_preserve_phase(
            complex_gains=bandpass[time], ref_ant=ref_ant
        )
        for pol in (0, 3):
            logger.info(f"Processing {pols[pol]} polarisation")

            for ant in range(solutions.nant):
                if ant == ref_ant:
                    logger.info(f"Skipping reference antenna = ant{ref_ant:02}")
                    continue

                ant_gains = ref_bandpass[ant, :, pol]
                plot_title = f"{title} - ant{ant:02d} - {pols[pol]}"
                output_path = (
                    plot_dir / f"{title}.ant{ant:02d}.{pols[pol]}.png"
                    if plot_dir is not None
                    else None
                )

                if np.sum(np.isfinite(ant_gains)) == 0:
                    logger.info(f"Not valid data found for ant{ant:0d} {pols[pol]}")
                    continue

                try:
                    phase_outlier_result = flag_outlier_phase(
                        complex_gains=ant_gains,
                        flag_cut=flag_cut,
                        plot_title=plot_title,
                        plot_path=output_path,
                    )
                    bandpass[time, ant, phase_outlier_result.outlier_mask, :] = np.nan
                except PhaseOutlierFitError:
                    # This is raised if the fit failed to converge, or some other nasty.
                    bandpass[time, ant, :, :] = np.nan

    for time in range(solutions.nsol):
        for pol in (0, 3):
            for ant in range(solutions.nant):
                # Flag all solutions for this (ant,pol) if more than 80% are flagged
                if flags_over_threshold(
                    flags=~np.isfinite(bandpass[time, ant, :, pol]),
                    thresh=0.8,
                    ant_idx=ant,
                ):
                    logger.info(
                        f"Flagging all solutions across  ant{ant:02d}, too many flagged channels."
                    )
                    bandpass[time, ant, :, :] = np.nan

                complex_gains = bandpass[time, ant, :, pol]
                if flag_mean_residual_amplitude(complex_gains=complex_gains):
                    logger.info(
                        f"Flagging all solutions for ant{ant:02d}, mean residual amplitudes high"
                    )
                    bandpass[time, ant, :, :] = np.nan

                flagged = ~np.isfinite(bandpass[time, ant, :, pol])
                logger.info(
                    f"{ant=:02d}, pol={pols[pol]}, flagged {np.sum(flagged) / ant_gains.shape[0] * 100.0:.2f}%"
                )

    for time in range(solutions.nsol):
        bandpass_phased_referenced = divide_bandpass_by_ref_ant_preserve_phase(
            complex_gains=bandpass[time], ref_ant=ref_ant
        )
        # This loop will flag based on stats across different polarisations
        for ant in range(solutions.nant):
            ant_gains = bandpass_phased_referenced[ant]
            if flag_mean_xxyy_amplitude_ratio(
                xx_complex_gains=ant_gains[:, 0],
                yy_complex_gains=ant_gains[:, 3],
                tolerance=mean_ant_tolerance,
            ):
                logger.info(f"{ant=} failed mean amplitude gain test. Flagging {ant=}.")
                bandpass[time, ant, :, :] = np.nan

    # To this point operations carried out to the bandpass were to the mutable array reference
    # so there is no need to create a new solutions instance
    out_solutions_path = get_aocalibrate_output_path(
        ms_path=solutions_path, include_preflagger=True, include_smoother=False
    )
    solutions.save(output_path=out_solutions_path)
    if plot_solutions_throughout:
        output_plots = plot_solutions(solutions=out_solutions_path, ref_ant=ref_ant)
        plots.extend(output_plots)

    if smooth_solutions:
        logger.info("Smoothing the bandpass solutions. ")
        for time in range(solutions.nsol):
            complex_gains = divide_bandpass_by_ref_ant_preserve_phase(
                complex_gains=bandpass[time], ref_ant=ref_ant
            )
            bandpass[time] = smooth_bandpass_complex_gains(
                complex_gains=complex_gains,
                window_size=smooth_window_size,
                polynomial_order=smooth_polynomial_order,
            )

        out_solutions_path = get_aocalibrate_output_path(
            ms_path=solutions_path, include_preflagger=True, include_smoother=True
        )
        solutions.save(output_path=out_solutions_path)
        if plot_solutions_throughout:
            output_plots = plot_solutions(solutions=out_solutions_path, ref_ant=None)
            plots.extend(output_plots)

    total_flagged = np.sum(~np.isfinite(bandpass)) / np.prod(bandpass.shape)
    if total_flagged > 0.8:
        msg = (
            f"{total_flagged * 100.0:.2f}% of {(solutions_path)!s} is flagged after running the preflagger. "
            "That is over 90%. "
            f"This surely can not be correct. Likely something has gone very wrong. "
        )
        logger.critical(msg)
        raise ValueError(msg)

    flagged_aosolutions = FlaggedAOSolution(
        path=out_solutions_path, plots=tuple(plots), bandpass=bandpass
    )

    return flagged_aosolutions


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="Run calibrate and apply the solutions given a measurement set and sky-model."
    )

    subparsers = parser.add_subparsers(
        dest="mode", help="AO Calibrate related operations"
    )
    calibrate_parser = subparsers.add_parser(
        "calibrate",
        help="Calibrate a MS using a text-based sky-model using AO calibrate",
    )

    calibrate_parser.add_argument(
        "ms",
        type=Path,
        help="The measurement set to calibrate and apply solutions to. ",
    )
    calibrate_parser.add_argument(
        "aoskymodel",
        type=Path,
        help="The AO-style sky-model file to use when calibrating. ",
    )
    calibrate_parser.add_argument(
        "--calibrate-container",
        type=Path,
        default="./calibrate.sif",
        help="The container containing calibrate and applysolutions. ",
    )
    calibrate_parser.add_argument(
        "--data-column", type=str, default="DATA", help="The column to calibrate"
    )

    apply_parser = subparsers.add_parser(
        "apply",
        help="Apply an existing AO-style solutions binary to a measurement set. ",
    )

    apply_parser.add_argument(
        "ms", type=Path, help="Path to the measurement set to apply the solutions to. "
    )
    apply_parser.add_argument(
        "aosolutions", type=Path, help="Path to the AO-style binary solutions file. "
    )
    apply_parser.add_argument(
        "--calibrate-container",
        type=Path,
        default="./calibrate.sif",
        help="The container containing calibrate and applysolutions. ",
    )
    apply_parser.add_argument(
        "--data-column", type=str, default="DATA", help="The column to calibrate"
    )

    flag_sols_parser = subparsers.add_parser(
        "flag",
        help="Attempt to flag the bandpass solutions in an ao-style binary solutions file",
    )

    flag_sols_parser.add_argument(
        "aosolutions", type=Path, help="Path to the solution file to inspect and flag"
    )
    flag_sols_parser.add_argument(
        "--flag-cut",
        type=float,
        default=3.0,
        help="The significance level that an outlier phase has to be before being flagged",
    )
    flag_sols_parser.add_argument(
        "--plot-dir",
        type=Path,
        default=None,
        help="Directory to write diagnostic plots to. If unset no plots will be created. ",
    )

    return parser


def cli() -> None:
    import logging

    parser = get_parser()

    args = parser.parse_args()

    logger.setLevel(logging.DEBUG)

    if args.mode == "calibrate":
        calibrate_apply_ms(
            ms_path=args.ms,
            model_path=args.aoskymodel,
            container=args.calibrate_container,
            data_column=args.data_column,
        )
    elif args.mode == "apply":
        apply_solutions_to_ms(
            ms=args.ms,
            solutions_path=args.aosolutions,
            container=args.calibrate_container,
            data_column=args.data_column,
        )
    elif args.mode == "flag":
        flag_aosolutions(
            solutions_path=args.aosolutions,
            flag_cut=args.flag_cut,
            plot_dir=args.plot_dir,
        )


if __name__ == "__main__":
    cli()



"""Utility functions to carry out flagging against ASKAP measurement sets"""

from __future__ import annotations

import math
from argparse import ArgumentParser
from pathlib import Path
from typing import Collection, NamedTuple

import numpy as np
from casacore.tables import table
from numpy.typing import NDArray

from flint.exceptions import MSError
from flint.logging import logger
from flint.ms import check_column_in_ms, critical_ms_interaction, describe_ms
from flint.options import MS
from flint.sclient import run_singularity_command
from flint.utils import get_packaged_resource_path


class AOFlaggerCommand(NamedTuple):
    """The command to use when running aoflagger"""

    cmd: str
    """The command that will be executed"""
    ms_path: Path
    """The path to the MS that will be flagged. """
    ms: MS
    """The MS object that was flagged"""
    strategy_file: Path | None = None
    """The path to the aoflagging strategy file to use"""


def flag_ms_zero_uvws(ms: MS, chunk_size: int = 10000) -> MS:
    """Flag out the UVWs in a measurement set that have values of zero.
    This happens when some data are flagged before it reaches the TOS.

    A critical MS interaction scope is created to ensure if things fail
    they are known.

    Args:
        ms (MS): Measurement set to flag
        chunk_size (int, optional): The number of rows to flag at a tim. Defaults to 10000.

    Returns:
        MS: The flagged measurement set
    """

    ms = MS.cast(ms)
    logger.info(f"Flagging zero uvw's for {ms.path}")
    row_idx = 0

    # Rename the measurement set while it is being operated on
    with critical_ms_interaction(input_ms=ms.path) as critical_ms_path:
        with table(str(critical_ms_path), readonly=False, ack=False) as tab:
            table_size = len(tab)

            # so long as the row index is less than the table size there
            # is another chunk to flag
            while row_idx < (table_size - 1):
                uvws = tab.getcol("UVW", startrow=row_idx, nrow=chunk_size)
                flags = tab.getcol("FLAG", startrow=row_idx, nrow=chunk_size)

                # Select records what the (u,v,w) are (0,0,0)
                # Data in the shape (record, 3)
                zero_uvws = np.all(uvws == 0, axis=1)
                flags[zero_uvws, :] = True

                # Put it back into place, update the counter for the next insertion
                size = len(flags)
                tab.putcol("FLAG", flags, startrow=row_idx, nrow=size)
                row_idx += size

                # Ensure changes written back to the MS
                tab.flush()

    return ms


def _nan_zero_extreme_flag(
    data: NDArray[np.complex128],
    flags: NDArray[np.bool],
    uvws: NDArray[np.float64],
    flag_extreme_dxy: bool = True,
    dxy_thresh: float = 4.0,
) -> tuple[NDArray[np.complex128], NDArray[np.bool], NDArray[np.float64]]:
    """Used to evaluate the flags that should be applied to the MS based on
    data values (if there are 0's or nans), uvw values (which can be 0's) and
    extreme stokes-v values.

    This is the internal checker for the larger `nan_zero_extreme_flags_in_ms`
    function.

    Args:
        data (NDArray[np.complex128]): Data that is being consider drawn from a `DATA` like column
        flags (NDArray[np.bool]): The flags that correspond to the `DATA` column drawn from the `FLAG` column
        uvws (NDArray[np.float64]): The (U,V,W) coordinates of the data
        flag_extreme_dxy (bool, optional): Whether flagging on extreme stokes-v should be consider. Defaults to True.
        dxy_thresh (float, optional): The threshold of the stokes-v flagging. Defaults to 4.0.

    Returns:
        tuple[NDArray[np.complex128], NDArray[np.bool], NDArray[np.float64]]: Copies of the input `data`, `flags` and `uvws`. The `flags` are updated based on the results of the flagging.
    """

    no_flags_before = np.sum(flags)

    nan_mask = np.where(~np.isfinite(data))
    zero_mask = np.where(data == 0 + 0j)
    uvw_mask = np.any(uvws == 0, axis=1)
    logger.info(
        f"Will flag {np.sum(nan_mask)} NaNs, zero'd data {np.sum(zero_mask)}, zero'd UVW {np.sum(uvw_mask)}. "
    )

    flags[nan_mask] = True
    flags[zero_mask] = True
    flags[uvw_mask] = True

    if flag_extreme_dxy:
        logger.info(f"Flagging based on extreme Stokes-V, threshold {dxy_thresh=}")
        dxy = np.abs(data[:, :, 1] - data[:, :, 2])
        dxy_mask = np.where(dxy > dxy_thresh)
        logger.info(f"Will flag {np.sum(dxy_mask)} extreme Stokes-V.")

        # TODO: This can be compressed to a one-liner
        flags[:, :, 0][dxy_mask] = True
        flags[:, :, 1][dxy_mask] = True
        flags[:, :, 2][dxy_mask] = True
        flags[:, :, 3][dxy_mask] = True

    no_flags_after = np.sum(flags)
    logger.info(
        f"Flags before: {no_flags_before}, Flags after: {no_flags_after}, Difference {no_flags_after - no_flags_before}"
    )

    return (data, flags, uvws)


def nan_zero_extreme_flag_ms(
    ms: Path | MS,
    data_column: str | None = None,
    flag_extreme_dxy: bool = True,
    dxy_thresh: float = 4.0,
    nan_data_on_flag: bool = False,
    chunk_size: int = 100000,
) -> MS:
    """Will flag a MS based on NaNs or zeros in the nominated data column of a measurement set.
    These NaNs might be introduced into a column via the application of a applysolutions task.
    Zeros might be introduced by the correlator dropping cycles and not appropriately setting the
    corresponding FLAG column (although this might have been fixed).

    There is also an optional component to flag based on extreme Stokes-V values.

    Visibilities that are marked as bad will have the FLAG column updatede appropriately.

    Args:
        ms (Union[Path,MS]): The measurement set that will be processed and have visibilities flagged.
        data_column (Optional[str], optional): The column to inspect. This will override the value in the nominated column of the MS. Defaults to None.
        flag_extreme_dxy (bool, optional): Whether Stokes-V will be inspected and flagged. Defaults to True.
        dxy_thresh (float, optional): Threshold used in the Stokes-V case. Defaults to 4.0.
        nan_data_on_flag (bool, optional): If True, data whose FLAG is set to True will become NaNs. Defaults to False.
        chunk_size (int, optional): Number of rows to process at a time. Defaults to 100000.

    Returns:
        MS: The container of the processed MS
    """
    ms = MS.cast(ms)

    if data_column is None and ms.column is None:
        logger.warning("No valid data column selected, using default of DATA")
        data_column = "DATA"
    elif data_column is None and ms.column is not None:
        logger.info(f"Using nominated {ms.column} column for {ms.path!s}")
        data_column = ms.column

    logger.info(f"Flagging NaNs and zeros in {data_column}.")

    with table(str(ms.path), readonly=False, ack=False) as tab:
        no_rows = len(tab)
        no_chunks = math.ceil(no_rows / chunk_size)
        logger.info(f"Number of rows: {no_rows} across {no_chunks}")
        idx = 0
        while idx * chunk_size < no_rows:
            start_row = idx * chunk_size
            logger.info(
                f"Flagging {idx + 1} of {no_chunks} - {start_row=} with {chunk_size=}"
            )

            data = tab.getcol(data_column, startrow=start_row, nrow=chunk_size)
            flags = tab.getcol("FLAG", startrow=start_row, nrow=chunk_size)
            uvws = tab.getcol("UVW", startrow=start_row, nrow=chunk_size)

            data, flags, uvws = _nan_zero_extreme_flag(
                data=data,
                flags=flags,
                uvws=uvws,
                flag_extreme_dxy=flag_extreme_dxy,
                dxy_thresh=dxy_thresh,
            )

            logger.info("Adding updated flags column")
            tab.putcol("FLAG", flags, startrow=start_row, nrow=chunk_size)

            # Keep this outside the above function to avoid
            # passing the table reference. To avoid thrashing the data
            # column wwe would otherwise check to see if this option is enabled
            # when putting the column. May as well keep it here.
            if nan_data_on_flag:
                data[flags] = np.nan
                logger.info(f"Setting {np.sum(flags)} DATA items to NaN.")
                tab.putcol(data_column, data, startrow=start_row, nrow=chunk_size)
            tab.flush()
            idx += 1

        tab.close()

    return ms


def create_aoflagger_cmd(ms: MS) -> AOFlaggerCommand:
    """Create a command to run aoflagger against a measurement set

    Args:
        ms (MS): The measurement set to flag. The column attached to the MS.column
        is flagged.

    Raises:
        MSError: Raised when the attached column is not found in the MS

    Returns:
        AOFlaggerCommand: The aoflagger command that will be run
    """
    logger.info("Creating an AOFlagger command. ")
    ms = MS.cast(ms)

    assert ms.column is not None, (
        f"MS column must be set in order to flag, currently {ms.column=}. Full {ms=}"
    )

    if not check_column_in_ms(ms):
        raise MSError(f"Column {ms.column} not found in {ms.path}.")

    flagging_strategy = get_packaged_resource_path(
        package="flint.data.aoflagger", filename="ASKAP.lua"
    )
    logger.info(f"Flagging using the strategy file {flagging_strategy}")

    cmd = f"aoflagger -column {ms.column} -strategy {flagging_strategy} -v {ms.path.absolute()!s}"

    return AOFlaggerCommand(
        cmd=cmd, ms_path=ms.path, strategy_file=Path(flagging_strategy), ms=ms
    )


def run_aoflagger_cmd(aoflagger_cmd: AOFlaggerCommand, container: Path) -> None:
    """Run the aoflagger command constructed in its singularity container

    Args:
        aoflagger_cmd (AOFlaggerCommand): The command that will be executed
        container (Path): Path to the container that contains aoflagger
    """
    assert container.exists(), (
        f"The applysolutions container {container} does not exist. "
    )

    bind_dirs = [aoflagger_cmd.ms_path.parent.absolute()]
    logger.debug(f"Bind directory for aoflagger: {bind_dirs}")

    if aoflagger_cmd.strategy_file:
        bind_dirs.append(aoflagger_cmd.strategy_file)

    run_singularity_command(
        image=container.absolute(), command=aoflagger_cmd.cmd, bind_dirs=bind_dirs
    )


def flag_ms_aoflagger(ms: MS, container: Path) -> MS:
    """Create and run an aoflagger command in a container

    Args:
        ms (MS): The measurement set with nominated column to flag
        container (Path): The container with the aoflagger program

    Returns:
        MS: Measurement set flagged with the appropriate column
    """
    ms = MS.cast(ms)
    logger.info(f"Will flag column {ms.column} in {ms.path!s}.")
    aoflagger_cmd = create_aoflagger_cmd(ms=ms)

    logger.info("Flagging command constructed. ")
    run_aoflagger_cmd(aoflagger_cmd=aoflagger_cmd, container=container)

    # TODO: This should be moved to the aoflagger lua file once it has
    # been implemented
    ms = flag_ms_zero_uvws(ms=ms)

    return ms


def flag_ms_by_antenna_ids(ms: Path | MS, ant_ids: int | Collection[int]) -> MS:
    """Set the FLAG to True for a collection of rows where ANTENNA1 or ANTENNA2 is in a set of
    antenna IDs to flag. The flagging is performed via the antenna ID as it is in the measurement
    set - it is not by the antenna name.

    Args:
        ms (Union[Path, MS]): The measurement set that has antennas to flag
        ant_ids (Union[int,Collection[int]]): The set of antenna IDs to flag.

    Returns:
        MS: The measurement set with flagged antennas.
    """
    ms = MS.cast(ms)

    ant_ids = (ant_ids,) if isinstance(ant_ids, int) else ant_ids

    if len(ant_ids) == 0:
        logger.info("Antenna list to flag is empty. Exiting. ")
        return ms

    logger.info(f"Will flag {ms.path!s}.")
    logger.info(f"Antennas to flag: {ant_ids}")

    # TODO: Potentially this should be batched into chunks to operate over
    with table(str(ms.path), readonly=False, ack=False) as tab:
        logger.info(f"Opened {ms.path!s}, loading metadata.")
        ant1 = tab.getcol("ANTENNA1")
        ant2 = tab.getcol("ANTENNA2")
        flags = tab.getcol("FLAG")

        init_flags = np.sum(flags)

        for ant_id in ant_ids:
            ant_mask = (ant_id == ant1) | (ant_id == ant2)

            if not np.any(ant_mask):
                logger.info(f"No data for {ant_id=} found. Continuing. ")
                continue

            logger.info(f"Flagging {ant_id=}.")
            flags[ant_mask] = True

        tab.putcol("FLAG", flags)
        end_flags = np.sum(flags)

    diff_flags = end_flags - init_flags
    logger.info(
        f"Loaded flags: {init_flags}, Final flags: {end_flags}, Difference: {diff_flags} ({diff_flags / np.prod(flags.shape) * 100.0:.2f}%)"
    )

    return ms


def get_parser() -> ArgumentParser:
    """Create the argument parser for the flagging

    Returns:
        ArgumentParser: aoflagger argument parser
    """
    parser = ArgumentParser(description="Run aoflagger against a measurement set")

    subparser = parser.add_subparsers(dest="mode")

    aoflagger_parser = subparser.add_parser(
        "aoflagger", description="Run AOFlagger against an measurement set. "
    )
    aoflagger_parser.add_argument("ms", type=Path, help="The measurement set to flag")
    aoflagger_parser.add_argument(
        "--aoflagger-container",
        type=Path,
        default=Path("aoflagger.sif"),
        help="The container that holds the aoflagger application",
    )
    aoflagger_parser.add_argument(
        "--column", type=str, default="DATA", help="The column of data in MS to flag"
    )

    nan_zero_parser = subparser.add_parser(
        "nanflag", help="Flag visibilities that are either NaN or zeros. "
    )
    nan_zero_parser.add_argument(
        "ms", type=Path, help="The measurement set that will be flagged. "
    )
    nan_zero_parser.add_argument(
        "--column", type=str, default="DATA", help="The column of data in MS to flag"
    )
    nan_zero_parser.add_argument(
        "--flag-extreme-dxy",
        action="store_true",
        help="Flag visibilities whose ABS(XY - YX) change significantly",
    )
    nan_zero_parser.add_argument(
        "--dxy-thresh",
        type=float,
        default=4.0,
        help="If extreme dxy flagging, ABS(XY - YX) above this value will be flagged. ",
    )
    nan_zero_parser.add_argument(
        "--nan-data-on-flag",
        action="store_true",
        help="NaN the data if their FLAG attribute is True. ",
    )

    antenna_parser = subparser.add_parser("antenna", help="Flag data by the antenna ID")
    antenna_parser.add_argument(
        "ms", type=Path, help="Path to the measurement set that will be flagged. "
    )
    antenna_parser.add_argument(
        "antenna_ids",
        type=int,
        nargs="+",
        help="The antenna IDs of the rows that should be flagged. ",
    )
    return parser


def cli() -> None:
    import logging

    logger.setLevel(logging.INFO)

    parser = get_parser()

    args = parser.parse_args()

    if args.mode == "aoflagger":
        ms = MS(path=args.ms, column=args.column)

        describe_ms(ms)
        flag_ms_aoflagger(ms=ms, container=args.aoflagger_container)
        describe_ms(ms)
    elif args.mode == "nanflag":
        nan_zero_extreme_flag_ms(
            ms=args.ms,
            data_column=args.column,
            flag_extreme_dxy=args.flag_extreme_dxy,
            dxy_thresh=args.dxy_thresh,
            nan_data_on_flag=args.nan_data_on_flag,
        )
    elif args.mode == "antenna":
        flag_ms_by_antenna_ids(ms=args.ms, ant_ids=args.antenna_ids)


if __name__ == "__main__":
    cli()



"""Utilities around catalogues.

Known reference catalogues are described by their ViZeR catalogue id,
which are used to download and store the appropriately formed catalogues on disk.
If the ViZeR service is down then attempts to download and form FITS catalgoues
will fail. These only need to be downloaded once, provided they can be stored
and retained on disk.
"""

from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path
from typing import NamedTuple

import astropy.units as u
from astropy.table import Table
from astroquery.vizier import Vizier

from flint.logging import logger


class Catalogue(NamedTuple):
    """A basic structure used to describe a known catalogue."""

    survey: str
    """Shorthand name of the sourcey catalogue"""
    file_name: str
    """The file name of the known catalogue"""
    freq: float  # Hertz
    """Reference frequency of the catalogue, in Hertz"""
    ra_col: str
    """Column name containing the right-ascension"""
    dec_col: str
    """Column name containing the declination"""
    name_col: str
    """Column name containing the source/component name"""
    flux_col: str
    """Column name containing the flux density"""
    maj_col: str
    """Column name containing the major-axis of the source gaussian component"""
    min_col: str
    """Column name containing the min-axis of the source gaussian component"""
    pa_col: str
    """Column name containing the pa of the source gaussian component"""
    alpha_col: str | None = None  # Used to scale the SED
    """Column name containing the spectral index, used to calculate the source SED. If None a default is used. """
    q_col: str | None = None  # Used to scale the SED
    """Column name containing the curvature of the spectral index, used to calculate the source SED. If None a default is used. """
    vizier_id: str | None = (
        None  # Required for known reference catalogues, not for other specified catalogues
    )
    """The ID of the catalogue on Vizier that is used to download the catalogue"""


KNOWN_REFERENCE_CATALOGUES: dict[str, Catalogue] = dict(
    NVSS=Catalogue(
        survey="NVSS",
        file_name="NVSS.fits",
        name_col="NVSS",
        freq=1.4e9,
        ra_col="_RAJ2000",
        dec_col="_DEJ2000",
        flux_col="S1.4",
        maj_col="MajAxis",
        min_col="MinAxis",
        pa_col="PA",
        vizier_id="VIII/65/nvss",
    ),
    SUMSS=Catalogue(
        survey="SUMSS",
        file_name="SUMSS.fits",
        freq=8.43e8,
        ra_col="_RAJ2000",
        dec_col="_DEJ2000",
        name_col="Mosaic",
        flux_col="St",
        maj_col="dMajAxis",
        min_col="dMinAxis",
        pa_col="dPA",
        vizier_id="VIII/81B/sumss212",
    ),
    ICRF=Catalogue(
        survey="ICRF",
        file_name="ICRF.fits",
        freq=1e9,
        ra_col="_RAJ2000",
        dec_col="_DEJ2000",
        name_col="ICRF",
        flux_col="None",
        maj_col="None",
        min_col="None",
        pa_col="None",
        vizier_id="I/323/icrf2",
    ),
    RACSLOW=Catalogue(
        file_name="racs-low.fits",
        survey="RACS-LOW",
        freq=887.56e6,
        ra_col="_RAJ2000",
        dec_col="_DEJ2000",
        name_col="GID",
        flux_col="Ftot",
        maj_col="amaj",
        min_col="bmin",
        pa_col="PA",
        vizier_id="J/other/PASA/38.58/gausscut",
    ),
)

# Helper functions that are used to try to guess column names from a table.
# These are intended to only be helpers for common names and not an extensive list
PREFERRED_RA_COLUMN_NAMES = ["RAJ2000", "ra"]
PREFERRED_DEC_COLUMN_NAMES = ["DEJ2000", "dec"]
PREFERRED_PEAK_COLUMN_NAMES = ["peak_flux", "Sp"]
PREFERRED_INT_COLUMN_NAMES = ["int_flux", "Sint"]
PREFERRED_INT_ERR_COLUMN_NAMES = ["local_rms"]

PREFERRED_COLUMNS = dict(
    ra=PREFERRED_RA_COLUMN_NAMES,
    dec=PREFERRED_DEC_COLUMN_NAMES,
    peakflux=PREFERRED_PEAK_COLUMN_NAMES,
    intflux=PREFERRED_INT_COLUMN_NAMES,
    intfluxerr=PREFERRED_INT_ERR_COLUMN_NAMES,
)


def guess_column_in_table(
    table: Table, column: str, guess_column: str | None = None
) -> str:
    """Attempt to deduce the appropriate column name from a set of
    column names in a table. A lookup of known column names for different
    contexts if consulted. Available modes are:

    #. ra
    #. dec
    #. peakflux
    #. intflux

    If `guess_column` is provided and is in the table this is returned

    Args:
        table (Table): The table with the column names to inspect
        column (str): The type of the column we are attempting to deduce.
        guess_column (Optional[str], optional): Consider whether this column exists first. Defaults to None.

    Raises:
        ValueError: Raised when either the RA or Dec columns could not be figured out

    Returns:
        str: The names of the peak flux column
    """

    logger.debug(f"Guessing column name for {column=} with {guess_column=}")
    column_names = [col.upper() for col in table.colnames]
    preferred_columns = PREFERRED_COLUMNS.get(column, None)
    if preferred_columns is None:
        raise KeyError(f"{column=} not in {PREFERRED_COLUMNS.keys()}")

    cols = (
        [guess_column]
        if guess_column and guess_column.upper() in column_names
        else [col for col in preferred_columns if col.upper() in column_names]
    )

    if not len(cols) > 0:
        raise ValueError(
            f"Unable to guess {column=} column names. Table has {column_names=}, and {preferred_columns=}"
        )

    return cols[0]


def _guess_catalogue_type(
    table: Table | Path,
    survey: str = "askap",
    file_name: str = "askap.fit",
    freq: float = -1,
) -> Catalogue:
    """This is a stub function that will be expanded in the future. It
    is intended to guess the appropriate source finder that constructed
    a component catalogue, resolving issues around the column names.

    This stub function will return a aegean catalogue description

    Args:
        table (Union[Table, Path]): The table that will be inspected
        survey (str, optional): The name of the survey to use. Defaults to askap.
        file_name (str, optional): The file name that would be used for the catalogue. Defaults of askap.fits.

    Returns:
        Catalogue: Placeholder Catalogue representing Aegean catalogues
    """
    table = Table.read(table) if isinstance(table, Path) else table

    return Catalogue(
        survey=survey,
        file_name=file_name,
        freq=freq,
        name_col="source",
        ra_col="ra",
        dec_col="dec",
        flux_col="int_flux",
        maj_col="a",
        min_col="b",
        pa_col="pa",
    )


def get_reference_catalogue(
    reference_directory: Path, survey: str, verify: bool = True
) -> tuple[Table, Catalogue]:
    """Load in a known reference catalogue

    If `verify` is `True` then the `Catalogue.flux_col`, if present, will have
    its units converted to Jy.

    Args:
        reference_directory (Path): The path to the directory where reference catalogues were downloaded to
        survey (str): The name of the survey to load.
        verify (bool, optional): If `True`, the table column names are inspected to ensure they are correct. Defaults to True.

    Raises:
        ValueError: Raised when the requested survey is not known

    Returns:
        Tuple[Table, Catalogue]: The loaded table and corresponding set of expected columns
    """

    catalogue = KNOWN_REFERENCE_CATALOGUES.get(survey, None)

    if catalogue is None:
        raise ValueError(
            f"{survey=} is not known, recognized reference catalogues are {KNOWN_REFERENCE_CATALOGUES.keys()}"
        )

    table_path = reference_directory / catalogue.file_name
    logger.info(f"Loading {table_path=}")

    if not table_path.exists():
        raise FileNotFoundError(
            f"{table_path=} not found. Check {reference_directory=} for known catalogues"
        )

    table = Table.read(table_path)

    if verify:
        valid_cols = [
            col in table.columns
            for col in (
                catalogue.ra_col,
                catalogue.dec_col,
                catalogue.name_col,
                catalogue.flux_col,
                catalogue.maj_col,
                catalogue.min_col,
            )
            if col.lower() != "none"
        ]
        assert all(valid_cols), f"Column is not valid, {valid_cols=}"
        if catalogue.flux_col.lower() != "none":
            assert isinstance(table[catalogue.flux_col].unit, u.Unit)
            table[catalogue.flux_col] = table[catalogue.flux_col].to(u.Jy)

    return table, catalogue


def download_vizier_catalogue(
    output_path: Path, vizier_id: str, dry_run: bool = False
) -> Path:
    """Download a catalogue from the vizier catalogue service. The table
    will be obtained using astroquery and written out to the supplied
    `output_path`, from which the format is inferred.

    Args:
        output_path (Path): Where the table will be written to
        vizier_id (str): The catalogue ID that will be downloaded
        dry_run (bool, optional): If `True`, no downloading will take place. Defaults to False.

    Returns:
        Path: Path the file was written to
    """
    logger.info(f"Downloading {vizier_id=}")

    if dry_run:
        logger.info(f"{dry_run=}, not downloading")
        return output_path

    tablelist = Vizier(
        columns=["_RAJ2000", "_DEJ2000", "all"], row_limit=-1
    ).get_catalogs(vizier_id, verbose=True)
    logger.info(f"catalogue downloaded, contains {len(tablelist[0])} rows")
    logger.info(f"Writing {vizier_id=} to {output_path=}")

    assert len(tablelist) == 1, (
        f"Table list for {vizier_id=} has unexpected length of {len(tablelist)}"
    )

    # Note all pirates respect the FITS standard@
    if description := tablelist[0].meta.get("description", None):
        tablelist[0].meta["description"] = description[:30]

    tablelist[0].write(output_path, overwrite=True)

    return output_path


def download_reference_catalogues(
    reference_directory: Path, dry_run: bool = False
) -> tuple[Path, ...]:
    """Download all of the expected reference catalogue data that flint relies on

    Args:
        reference_directory (Path): The parent directory catalgoues will be written to
        dry_run (bool, optional): If `True`, no downloading will take place. Defaults to False.

    Returns:
        Tuple[Path, ...]: Collection of paths of all the downloaded reference catalogues
    """

    logger.info(f"Downloading catalogues to {reference_directory=}")
    reference_directory.mkdir(parents=True, exist_ok=True)

    logger.info(f"Downloading {len(KNOWN_REFERENCE_CATALOGUES)} catalogues")
    catalogue_paths = [
        download_vizier_catalogue(
            output_path=(reference_directory / f"{catalogue.file_name}").absolute(),
            vizier_id=catalogue.vizier_id,
            dry_run=dry_run,
        )
        for _, catalogue in KNOWN_REFERENCE_CATALOGUES.items()
        if catalogue.vizier_id
    ]

    return tuple(catalogue_paths)


def list_known_reference_catalogues() -> None:
    """List the known reference catalogues that are expected/downloaded by flint"""

    logger.info(f"{len(KNOWN_REFERENCE_CATALOGUES)} are known")
    for survey, cata in KNOWN_REFERENCE_CATALOGUES.items():
        logger.info(f"{survey=}")
        logger.info(f"{Catalogue}")


def verify_reference_catalogues(
    reference_directory: Path, load_catalogue: bool = True
) -> bool:
    """Attempt to load the set of reference catalogues to ensure they are correctly
    formed

    Args:
        reference_directory (Path): The directory containing the reference catalogues
        load_catalogue (bool, optional): Load the catalogue as part of verification. If False only check to see if the file exists. Defaults to True.

    Returns:
        bool: Indicates whether all catalogue files exist and are correctly formed
    """

    logger.info(f"Verifying catalogues in {reference_directory=}")
    logger.info(f"Searching for {len(KNOWN_REFERENCE_CATALOGUES)}")
    survey_valid = {}
    for survey, cata in KNOWN_REFERENCE_CATALOGUES.items():
        try:
            if load_catalogue:
                _ = get_reference_catalogue(
                    reference_directory=reference_directory, survey=survey
                )
            else:
                cata_path = reference_directory / cata.file_name
                assert cata_path.exists()
            valid = True
        except (ValueError, AssertionError, FileNotFoundError):
            valid = False
        logger.info(f"{survey=} is {'valid' if valid else 'not valid'}")
        survey_valid[survey] = valid

    return all(survey_valid.values())


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(description="Utilities around catalogues")

    subparser = parser.add_subparsers(
        dest="mode", help="Operation mode of flint_catalogue"
    )

    download_parser = subparser.add_parser(
        "download", help="Download reference catalogues"
    )
    download_parser.add_argument(
        "reference_directory",
        type=Path,
        help="The directory to save the reference catalogues to",
    )

    _ = subparser.add_parser("list", help="List the known reference catalogues")

    verify_parser = subparser.add_parser(
        "verify",
        help="Ensure the expected catalogues existing in the reference directory and are correctly formed",
    )
    verify_parser.add_argument(
        "reference_directory",
        type=Path,
        help="Directory containing the known reference catalogues",
    )

    return parser


def cli() -> None:
    parser = get_parser()

    args = parser.parse_args()

    if args.mode == "download":
        download_reference_catalogues(reference_directory=args.reference_directory)
    elif args.mode == "list":
        list_known_reference_catalogues()
    elif args.mode == "verify":
        verify_reference_catalogues(reference_directory=args.reference_directory)
    else:
        logger.info(f"Mode {args.mode} is not recognised")
        parser.print_help()


if __name__ == "__main__":
    cli()



"""Items related to test functions in the validation stage of flint"""

from __future__ import annotations

from pathlib import Path
from typing import NamedTuple

import matplotlib.pyplot as plt
import numpy as np
import pytest
from astropy.table import Table

from flint.utils import get_packaged_resource_path
from flint.validation import (
    RMSImageInfo,
    SourceCounts,
    calculate_area_correction_per_flux,
    extract_inner_image_array_region,
    get_parser,
    get_rms_image_info,
    get_source_counts,
    plot_source_counts,
)


def test_get_parser():
    get_parser()


@pytest.fixture
def rms_path(tmpdir):
    rms_path = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB39400.RACS_0635-31.beam0-MFS-subimage_rms.fits",
        )
    )

    return rms_path


@pytest.fixture
def comp_path(tmpdir):
    comp_path = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB38959.RACS_1357-18.noselfcal.linmos_comp.fits",
        )
    )

    return comp_path


def test_extract_inner_image_array_region():
    """Get the inner region of an image array"""
    shape = (100, 100)
    image = np.arange(np.prod(shape)).reshape(shape)

    sub_image = extract_inner_image_array_region(image=image, fraction=0.5)
    assert sub_image.shape == (50, 50)

    shape = (1, 1, 100, 100)
    image = np.arange(np.prod(shape)).reshape(shape)

    sub_image = extract_inner_image_array_region(image=image, fraction=0.5)
    assert sub_image.shape == (1, 1, 50, 50)

    with pytest.raises(AssertionError):
        extract_inner_image_array_region(image=image, fraction=22.5)


def test_plot_source_counts(comp_path, rms_path):
    """Just a simple test to make sure there are no immediate matplotlib api errors.
    Not testing for actual plotting results."""
    fig, ax = plt.subplots(1, 1)
    comp_table = Table.read(comp_path)
    rms_info = get_rms_image_info(rms_path=rms_path)

    plot_source_counts(catalogue=comp_table, rms_info=rms_info, ax=ax)


def test_source_counts():
    fluxes = np.linspace(0.00003, 0.01, 10000)

    source_counts = get_source_counts(fluxes=fluxes, area=10)

    assert isinstance(source_counts, SourceCounts)
    assert source_counts.area == 10.0


def test_source_counts_with_rms(rms_path):
    fluxes = np.linspace(0.00003, 0.01, 10000)

    source_counts = get_source_counts(fluxes=fluxes, area=10, rms_image_path=rms_path)

    assert isinstance(source_counts, SourceCounts)
    assert source_counts.area == 10.0
    assert isinstance(source_counts.area_fraction, np.ndarray)


def test_calculate_area_correction(rms_path):
    """This is not an entirely robust check as the rms image is only
    10 x 15 pixels. More testing for errors in the calling."""

    flux_bin_centres = np.arange(0.00004, 0.01, 0.0001)

    area_frac = calculate_area_correction_per_flux(
        rms_image_path=rms_path, flux_bin_centre=flux_bin_centres, sigma=3
    )

    test = np.ones(100)
    test[:5] = 0.0

    assert np.allclose(test, area_frac)


def test_rms_image_info(rms_path):
    rms_info = get_rms_image_info(rms_path=rms_path)

    print(rms_info)

    assert isinstance(rms_info, RMSImageInfo)
    assert rms_info.path == rms_path
    assert rms_info.no_valid_pixels == 150
    assert rms_info.shape == (10, 15)
    assert np.isclose(0.0001515522, rms_info.median, atol=1e-5)
    assert np.isclose(0.00015135764, rms_info.minimum, atol=1e-5)
    assert np.isclose(0.0001518184, rms_info.maximum, atol=1e-5)
    assert np.isclose(1.1098655e-07, rms_info.std, atol=1e-5)


class Example(NamedTuple):
    a: int
    b: str
    c: float


def test_expected_namedtuple_get():
    """This is a simple test to ensure the behaviour of
    NamedTuple.__getattribute__ remains OK. This is currently
    used in the validation plotting to iterate over known
    surveys
    """

    test = Example(a=1, b="123", c=1.23)

    assert test.__getattribute__("a") == 1
    assert test.__getattribute__("b") == "123"
    assert test.__getattribute__("c") == 1.23



"""Bits around testing the convolution utilities"""

from __future__ import annotations

import shutil
from pathlib import Path

import numpy as np
import pytest
from astropy.io import fits

from flint.convol import (
    BeamShape,
    check_if_cube_fits,
    get_cube_common_beam,
)
from flint.utils import get_packaged_resource_path


@pytest.fixture
def image_fits() -> Path:
    image = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB39400.RACS_0635-31.beam0-MFS-subimage_rms.fits",
        )
    )

    return image


@pytest.fixture
def cube_fits(tmpdir) -> Path:
    """Extract some example cubes"""
    tmp_dir = Path(tmpdir)
    cube_dir = Path(tmp_dir / "cubes")
    cube_dir.mkdir(parents=True, exist_ok=True)

    cubes_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests", filename="sub_cube_fits_examples.zip"
        )
    )
    assert cubes_zip.exists()
    shutil.unpack_archive(cubes_zip, cube_dir)

    return cube_dir


def test_check_if_cube_fits(cube_fits, image_fits):
    """See if the cube fits checker is picking up cubes with 3 axis"""
    fits_files = list(cube_fits.glob("*sub.fits"))
    assert len(fits_files) == 10
    assert all([check_if_cube_fits(fits_file=f) for f in fits_files])

    assert not check_if_cube_fits(fits_file=image_fits)
    assert not check_if_cube_fits(fits_file=Path("ThisDoesNotExist"))


def test_get_cube_common_beam_and_convol_cubes(cube_fits) -> None:
    """Ensure that the common beam functionality of from beamcon_3D. Also test the
    convolution to the cubes, as the initial compute can be expensive"""
    fits_files = list(cube_fits.glob("*sub.fits"))
    assert len(fits_files) == 10

    data = fits.getdata(fits_files[0])
    data_shape = np.squeeze(data).shape  # type: ignore

    beam_list = get_cube_common_beam(cube_paths=fits_files, cutoff=150.0)
    assert len(beam_list) == data_shape[0]
    assert all([isinstance(b, BeamShape) for b in beam_list])


# This can cause thread locks in testing/ Test works. Test passes but
# produces something like the below when pytest wraps up
# =============== 257 passed, 14075 warnings in 526.16s (0:08:46) ================
# Fatal Python error: _enter_buffered_busy: could not acquire lock for <_io.BufferedWriter name='<stderr>'> at interpreter shutdown, possibly due to daemon threads
# Python
# def test_beam_list_convol(cube_fits):
#     # These come from the beam_list above
#     bmaj_arcsec = [
#         14.7,
#         14.5,
#         14.3,
#         14.2,
#         14.0,
#         13.8,
#         13.8,
#         13.5,
#         13.4,
#         32.9,
#         float("nan"),
#         147.5,
#         13.0,
#         12.9,
#         12.9,
#         12.8,
#         12.7,
#         13.0,
#         40.7,
#         12.7,
#         12.3,
#         12.3,
#         12.2,
#         12.1,
#         12.0,
#         11.9,
#         11.9,
#         11.6,
#         11.4,
#         11.6,
#         11.6,
#         11.5,
#         11.4,
#         11.4,
#         11.4,
#         11.3,
#     ]
#     bmin_arcsec = [
#         12.4,
#         12.2,
#         12.1,
#         12.0,
#         11.8,
#         11.7,
#         11.7,
#         11.6,
#         11.4,
#         16.2,
#         float("nan"),
#         63.7,
#         11.0,
#         10.9,
#         10.8,
#         10.7,
#         10.6,
#         10.8,
#         16.6,
#         10.6,
#         10.3,
#         10.2,
#         10.1,
#         10.0,
#         9.9,
#         9.9,
#         9.8,
#         9.7,
#         9.3,
#         9.6,
#         9.5,
#         9.4,
#         9.4,
#         9.3,
#         9.2,
#         9.2,
#     ]
#     bpa_deg = [
#         75.25,
#         75.02,
#         74.82,
#         74.82,
#         75.0,
#         75.01,
#         75.33,
#         76.03,
#         75.43,
#         -21.79,
#         float("nan"),
#         160.53,
#         73.28,
#         73.22,
#         76.36,
#         76.14,
#         76.07,
#         74.09,
#         55.61,
#         76.84,
#         75.9,
#         75.76,
#         75.96,
#         75.1,
#         75.76,
#         75.65,
#         75.66,
#         73.47,
#         71.63,
#         75.29,
#         75.35,
#         75.69,
#         75.11,
#         75.78,
#         75.84,
#         75.89,
#     ]

#     beam_list = [
#         BeamShape(bmaj_arcsec=bmaj, bmin_arcsec=bmin, bpa_deg=bpa)
#         for bmaj, bmin, bpa in zip(bmaj_arcsec, bmin_arcsec, bpa_deg)
#     ]

#     fits_files = list(cube_fits.glob("*sub.fits"))
#     assert len(fits_files) == 10
#     # This appears to make pytest lock up

#     cube_paths = convolve_cubes(
#         cube_paths=fits_files,
#         beam_shapes=beam_list,
#         cutoff=150.0,
#         executor_type="process",
#     )
#     assert all([isinstance(p, Path) for p in cube_paths])
#     assert all([p.exists() for p in cube_paths])



from __future__ import annotations

import filecmp
from pathlib import Path

import pytest

from flint.configuration import (
    Strategy,
    _create_mode_mapping_defaults,
    copy_and_timestamp_strategy_file,
    create_default_yaml,
    get_image_options_from_yaml,
    get_options_from_strategy,
    get_selfcal_options_from_yaml,
    load_strategy_yaml,
    verify_configuration,
    write_strategy_to_yaml,
)
from flint.utils import get_packaged_resource_path


@pytest.fixture
def package_strategy():
    example = get_packaged_resource_path(
        package="flint", filename="data/tests/test_config.yaml"
    )

    strategy = load_strategy_yaml(input_yaml=example, verify=False)

    return strategy


@pytest.fixture
def package_strategy_operations():
    example = get_packaged_resource_path(
        package="flint", filename="data/tests/test_config_2.yaml"
    )

    strategy = load_strategy_yaml(input_yaml=example, verify=False)

    return strategy


@pytest.fixture
def package_strategy_polarisation():
    example = get_packaged_resource_path(
        package="flint", filename="data/tests/test_config_pol.yaml"
    )

    strategy = load_strategy_yaml(input_yaml=example, verify=False)

    return strategy


@pytest.fixture
def strategy(tmpdir):
    output = create_default_yaml(
        output_yaml=Path(tmpdir) / "example.yaml", selfcal_rounds=3
    )

    strategy = load_strategy_yaml(input_yaml=output, verify=False)

    return strategy


@pytest.fixture
def package_strategy_path():
    example = get_packaged_resource_path(
        package="flint", filename="data/tests/test_config.yaml"
    )

    return example


def test_copy_and_timestamp(tmpdir):
    # a single function toe rename and copy a file because pirates needs to be efficient
    example = get_packaged_resource_path(
        package="flint", filename="data/tests/test_config.yaml"
    )
    copy_path = copy_and_timestamp_strategy_file(output_dir=tmpdir, input_yaml=example)

    assert copy_path != example
    assert filecmp.cmp(example, copy_path)


def test_verify_options_with_class_operations(package_strategy_operations):
    """Check whether the stokes-v and other operations around the
    verification and extraction of properties"""
    strategy = package_strategy_operations
    verify_configuration(input_strategy=strategy)

    strategy["ThisOperationDoesNotExists"] = {}
    with pytest.raises(ValueError):
        verify_configuration(input_strategy=strategy)

    strategy.pop("ThisOperationDoesNotExists")
    strategy["stokesv"]["wsclean"]["ThisDoesNotExist"] = "123"
    with pytest.raises(ValueError):
        verify_configuration(input_strategy=strategy)


def test_verify_getoptions_with_class_operations(package_strategy_operations):
    """Check whether the get options interface works with the operations
    section of the strategy file"""
    strategy = package_strategy_operations
    verify_configuration(input_strategy=strategy)

    options = get_options_from_strategy(
        strategy=package_strategy_operations, mode="wsclean", operation="stokesv"
    )
    assert options["pol"] == "V"
    assert options["channels_out"] == 2
    assert options["deconvolution_channels"] is None


def test_verify_options_with_class_missing_defaults(package_strategy):
    """Ensure that a error is raised if the defaults section is missing"""
    strategy = package_strategy
    verify_configuration(input_strategy=strategy)

    strategy.pop("defaults")
    with pytest.raises(ValueError):
        verify_configuration(input_strategy=strategy)


def test_verify_options_with_class(package_strategy):
    # ensure that the errors raised from options passed through
    # to the input structures correctly raise errors should they
    # be misconfigured (e.g. option supplied does not exist, missing
    # mandatory argument)
    strategy = package_strategy
    verify_configuration(input_strategy=strategy)

    strategy["selfcal"][0]["wsclean"]["ThisDoesNotExist"] = "ThisDoesNotExist"
    with pytest.raises(ValueError):
        verify_configuration(input_strategy=strategy)

    strategy["selfcal"][0]["wsclean"].pop("ThisDoesNotExist")
    verify_configuration(input_strategy=strategy)

    strategy["selfcal"][1]["masking"]["ThisDoesNotExist"] = "ThisDoesNotExist"

    with pytest.raises(ValueError):
        verify_configuration(input_strategy=strategy, raise_on_error=True)


def test_create_yaml_file(tmpdir):
    # ensure a default yaml strategy can be created
    output = create_default_yaml(
        output_yaml=Path(tmpdir) / "example.yaml", selfcal_rounds=3
    )

    assert output.exists()


def test_create_and_load(tmpdir):
    # ensure that a default strategy file can be both created and read back in
    output = create_default_yaml(
        output_yaml=Path(tmpdir) / "example.yaml", selfcal_rounds=3
    )

    assert output.exists()

    strategy = load_strategy_yaml(input_yaml=output)
    assert isinstance(strategy, Strategy)

    strategy = load_strategy_yaml(input_yaml=output, verify=False)
    assert isinstance(strategy, Strategy)


def test_verify(tmpdir):
    # test to make sure we can generate a default strategy (see pytest fixture)
    # read it backinto a dict and verify it is valid
    output = create_default_yaml(
        output_yaml=Path(tmpdir) / "example.yaml", selfcal_rounds=3
    )

    assert output.exists()
    strategy = load_strategy_yaml(input_yaml=output, verify=False)
    assert isinstance(strategy, Strategy)

    _ = verify_configuration(input_strategy=strategy)

    strategy["ddd"] = 123
    with pytest.raises(ValueError):
        verify_configuration(input_strategy=strategy)


def test_load_yaml_none():
    # make sure an error is raise if None is passed in. This
    # should be checked as the default value of FieldOptions.imaging_strategy
    # is None.
    with pytest.raises(TypeError):
        _ = load_strategy_yaml(input_yaml=None)  # type: ignore


def test_mode_options_mapping_creation():
    """Test that for each of the options in MODE_OPTIONS_MAPPING options can be default created"""
    defaults = _create_mode_mapping_defaults()
    assert isinstance(defaults, dict)


def test_get_options(strategy):
    # test to make sure we can generate a default strategy (see pytest fixture)
    # read it backinto a dict and then access some attributes
    wsclean = get_options_from_strategy(
        strategy=strategy, mode="wsclean", round_info=0, operation="selfcal"
    )
    assert isinstance(wsclean, dict)
    # example options
    assert wsclean["data_column"] == "CORRECTED_DATA"

    wsclean = get_options_from_strategy(
        strategy=strategy, mode="wsclean", round_info=1, operation="selfcal"
    )
    assert isinstance(wsclean, dict)
    # example options
    assert wsclean["data_column"] == "CORRECTED_DATA"

    archive = get_options_from_strategy(
        strategy=strategy, mode="archive", round_info=0, operation="selfcal"
    )
    assert isinstance(archive, dict)
    assert len(archive) > 0

    archive = get_options_from_strategy(
        strategy=strategy, mode="archive", operation="selfcal"
    )
    assert isinstance(archive, dict)
    assert len(archive) > 0


def test_get_mask_options(package_strategy):
    """Basic test to prove masking operation behaves well"""
    masking = get_options_from_strategy(
        strategy=package_strategy, mode="masking", round_info=0, operation="selfcal"
    )

    assert isinstance(masking, dict)
    assert masking["flood_fill_positive_seed_clip"] == 4.5

    masking2 = get_options_from_strategy(
        strategy=package_strategy, mode="masking", round_info=1, operation="selfcal"
    )

    print(strategy)

    assert isinstance(masking2, dict)
    assert masking2["flood_fill_positive_seed_clip"] == 40

    for k in masking.keys():
        if k == "flood_fill_positive_seed_clip":
            continue

        assert masking[k] == masking2[k]


def test_get_mask_options_from_path(package_strategy_path):
    """Basic test to prove masking operation behaves well"""
    package_strategy = Path(package_strategy_path)

    masking = get_options_from_strategy(
        strategy=package_strategy, mode="masking", round_info=0, operation="selfcal"
    )

    assert isinstance(masking, dict)
    assert masking["flood_fill_positive_seed_clip"] == 4.5

    masking2 = get_options_from_strategy(
        strategy=package_strategy, mode="masking", round_info=1, operation="selfcal"
    )

    assert isinstance(masking2, dict)
    assert masking2["flood_fill_positive_seed_clip"] == 40

    for k in masking.keys():
        if k == "flood_fill_positive_seed_clip":
            continue

        assert masking[k] == masking2[k]


def test_get_modes(package_strategy):
    # makes sure defaults for these modes are return when reuestion options
    # on a self-cal round without them set
    strategy = package_strategy

    for mode in ("wsclean", "gaincal", "masking"):
        test = get_options_from_strategy(
            strategy=strategy, mode=mode, round_info=1, operation="selfcal"
        )
        assert isinstance(test, dict)
        assert len(test.keys()) > 0


def test_bad_round(package_strategy):
    # make sure incorrect round is handled properly
    with pytest.raises(AssertionError):
        _ = get_options_from_strategy(
            strategy=package_strategy, round_info="doesnotexists", operation="selfcal"
        )

    with pytest.raises(AssertionError):
        _ = get_options_from_strategy(
            strategy=package_strategy, round_info=1.23456, operation="selfcal"
        )


def test_empty_strategy_empty_options():
    # if None is given as a strategy state then empty set of options is return
    res = get_options_from_strategy(strategy=None, operation="selfcal")

    assert isinstance(res, dict)
    assert not res == 0


def test_max_round_override(package_strategy):
    # ebsyre that the logic to switch to the highest available slefcal
    # round is sound
    strategy = package_strategy

    opts = get_options_from_strategy(
        strategy=strategy, round_info=9999, operation="selfcal"
    )
    assert isinstance(opts, dict)
    assert opts["data_column"] == "TheLastRoundIs3"


def test_empty_options(package_strategy):
    # ensures an empty options dict is return should something not set is requested
    items = get_options_from_strategy(
        strategy=package_strategy, mode="thisdoesnotexist", operation="selfcal"
    )

    assert isinstance(items, dict)
    assert len(items.keys()) == 0


def test_assert_strategy_bad():
    # tests to see if a error is raised when a bad strategy instance is passed in
    with pytest.raises(AssertionError):
        _ = get_options_from_strategy(strategy="ThisIsBad", operation="selfcal")


def test_updated_get_options(package_strategy):
    # test to make sure that the defaults outlined in a strategy file
    # are correctly overwritten should there be an options in a later
    # round. All other optiosn should remain the same.
    strategy = package_strategy
    assert isinstance(strategy, Strategy)

    wsclean_init = get_options_from_strategy(
        strategy=strategy, mode="wsclean", round_info=0, operation="selfcal"
    )
    assert wsclean_init["data_column"] == "CORRECTED_DATA"

    wsclean_1 = get_options_from_strategy(
        strategy=strategy, mode="wsclean", round_info=1, operation="selfcal"
    )
    assert wsclean_init["data_column"] == "CORRECTED_DATA"
    assert wsclean_1["data_column"] == "EXAMPLE"

    wsclean_2 = get_options_from_strategy(
        strategy=strategy, mode="wsclean", round_info=2, operation="selfcal"
    )
    assert wsclean_2["multiscale"] is False
    assert wsclean_2["data_column"] == "CORRECTED_DATA"

    assert all(
        [
            wsclean_init[key] == wsclean_1[key]
            for key in wsclean_init.keys()
            if key != "data_column"
        ]
    )
    assert wsclean_init["data_column"] != wsclean_1["data_column"]

    assert all(
        [
            wsclean_init[key] == wsclean_2[key]
            for key in wsclean_init.keys()
            if key != "multiscale"
        ]
    )


def test_get_image_options():
    init = get_image_options_from_yaml(input_yaml=None, self_cal_rounds=False)

    assert isinstance(init, dict)

    rounds = get_image_options_from_yaml(input_yaml=None, self_cal_rounds=True)

    assert isinstance(init, dict)
    assert 1 in rounds.keys()


def test_raise_image_options_error():
    example = Path("example.yaml")

    with pytest.raises(AssertionError):
        get_image_options_from_yaml(input_yaml=example)


def test_self_cal_options():
    rounds = get_selfcal_options_from_yaml(input_yaml=None)

    assert isinstance(rounds, dict)
    assert 1 in rounds.keys()


def test_raise_error_options_error():
    example = Path("example.yaml")

    with pytest.raises(AssertionError):
        get_selfcal_options_from_yaml(input_yaml=example)


def test_write_strategy_to_yaml(package_strategy, tmpdir):
    strategy = package_strategy

    output_strategy = Path(tmpdir) / "testing.yaml"
    write_strategy_to_yaml(strategy=strategy, output_path=output_strategy)

    loaded_strategy = load_strategy_yaml(input_yaml=output_strategy)

    assert len(set(loaded_strategy.keys()) - set(strategy.keys())) == 0
    assert (
        len(
            set(loaded_strategy["selfcal"][0].keys())
            - set(strategy["selfcal"][0].keys())
        )
        == 0
    )
    assert (
        len(
            set(loaded_strategy["selfcal"][1].keys())
            - set(strategy["selfcal"][1].keys())
        )
        == 0
    )
    assert (
        loaded_strategy["selfcal"][1]["wsclean"]["data_column"]
        == strategy["selfcal"][1]["wsclean"]["data_column"]
    )


def test_polarisation_options(package_strategy_polarisation):
    strategy = package_strategy_polarisation
    verify_configuration(input_strategy=strategy)

    total_options = get_options_from_strategy(
        strategy=strategy,
        operation="polarisation",
        polarisation="total",
    )
    assert total_options["pol"] == "i"
    assert not total_options["squared_channel_joining"]

    linear_options = get_options_from_strategy(
        strategy=strategy,
        operation="polarisation",
        polarisation="linear",
    )
    assert linear_options["pol"] == "qu"
    assert linear_options["squared_channel_joining"]
    assert linear_options["join_polarizations"]

    circular_options = get_options_from_strategy(
        strategy=strategy,
        operation="polarisation",
        polarisation="circular",
    )
    assert circular_options["pol"] == "v"



"""A basic interface into aegean source finding routines."""

from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path
from typing import Any, NamedTuple

from astropy.io import fits

from flint.exceptions import AttemptRerunException
from flint.logging import logger
from flint.naming import create_aegean_names
from flint.options import BaseOptions, add_options_to_parser, create_options_from_parser
from flint.sclient import run_singularity_command


class BANEOptions(BaseOptions):
    """Container for basic BANE related options. Only a subclass of BANE options are supported."""

    grid_size: tuple[int, int] | None = (16, 16)
    """The step interval of each box, in pixels"""
    box_size: tuple[int, int] | None = (196, 196)
    """The size of the box in pixels"""
    cores: int = 12
    """Number of cores to use. The number of stripes will be less than this number."""


class AegeanOptions(BaseOptions):
    """Container for basic aegean options. Only a subclass of aegean options are supported.

    Of note is the lack of a tables option (corresponding to --tables). This is dependent on knowing the base output name
    and relying on aegean to also append a suffix of sorts to the outputs. For that reason
    the aegean command generated will always create the table option.
    """

    nocov: bool = True
    """Whether aegean should attempt to model the co-variance of pixels. If true aegean does not. """
    maxsummits: int = 4
    """The maximum number of components an island is allowed to have before it is ignored. """
    autoload: bool = True
    """Attempt to load precomputed background and rms maps. """


class AegeanOutputs(NamedTuple):
    """Somple structure to represent output aegean products"""

    bkg: Path
    """Background map created by BANE"""
    rms: Path
    """RMS map created by BANE"""
    comp: Path
    """Source component catalogue created by Aegean"""
    beam_shape: tuple[float, float, float]
    """The `BMAJ`, `BMIN` and `BPA` that were stored in the image header that Aegen searched"""
    image: Path
    """The input image that was used to source find against"""


def _get_bane_command(image: Path, bane_options: BANEOptions) -> str:
    """Create the BANE command to run"""
    # The stripes is purposely set lower than the cores due to an outstanding bane bug that can cause a deadlock.
    cores = max(1, bane_options.cores)
    stripes = max(1, cores - 1)
    bane_command_str = f"BANE {image!s} --cores {cores} --stripes {stripes} "
    if bane_options.grid_size:
        bane_command_str += (
            f"--grid {bane_options.grid_size[0]} {bane_options.grid_size[1]} "
        )
    if bane_options.box_size:
        bane_command_str += (
            f"--box {bane_options.box_size[0]} {bane_options.box_size[1]}"
        )
    bane_command_str = bane_command_str.rstrip()
    logger.info("Constructed bane command.")

    return bane_command_str


def _bane_output_callback(line: str) -> None:
    """Callback handler for the BANE program. Will raise an error
    on the 'deadlock' issue."""

    assert isinstance(line, str)

    if "must be strictly ascending or descending" in line:
        logger.info("Potential BANE deadlock detectedc. Sleeping and raising error.")
        from time import sleep

        sleep(2)
        raise AttemptRerunException("BANE deadlock detected. ")


def _get_aegean_command(
    image: Path, base_output: str, aegean_options: AegeanOptions
) -> str:
    """Create the aegean command to run"""
    aegean_command = f"aegean {image!s} "
    if aegean_options.autoload:
        aegean_command += "--autoload "
    if aegean_options.nocov:
        aegean_command += "--nocov "

    # NOTE: Aegean will add the '_comp' component to the output tables. So, if we want
    # basename_comp.fits
    # to be the output component table then we want to pass
    # --table basename.fits
    # and have to rely on aegean doing the right thing.
    aegean_command += (
        f"--maxsummits {aegean_options.maxsummits} --table {base_output}.fits"
    )

    logger.info("Constructed aegean command. ")
    logger.debug(f"{aegean_command=}")

    return aegean_command


def run_bane_and_aegean(
    image: Path,
    aegean_container: Path,
    bane_options: BANEOptions | None = None,
    aegean_options: AegeanOptions | None = None,
    update_bane_options: dict[str, Any] | None = None,
    update_aegean_options: dict[str, Any] | None = None,
) -> AegeanOutputs:
    """Run BANE, the background and noise estimator, and aegean, the source finder,
    against an input image. This function attempts to hook into the AegeanTools
    module directly, which does not work with dask daemon processes.

    Args:
        image (Path): The input image that BANE will calculate a background and RMS map for
        aegean_container (Path): Path to a singularity container that was the AegeanTools packages installed.
        bane_options (Optional[BANEOptions], optional): The options that are provided to BANE. If None defaults of BANEOptions are used. Defaults to None.
        aegean_options (Optional[AegeanOptions], optional): The options that are provided to Aegean. if None defaults of AegeanOptions are used. Defaults to None.
        update_bane_options (dict[str, Any] | None, optional): Over-ride any default options of BANEOptions. If None defaults are used. Defaults to None.
        update_aegean_options (dict[str, Any] | None, optional): Over-ride any default options of AegeanOptions. If None defaults are used. Defaults to None.

    Returns:
        AegeanOutputs: The newly created BANE products
    """
    bane_options = bane_options if bane_options else BANEOptions()
    if update_bane_options:
        bane_options = bane_options.with_options(**update_bane_options)

    aegean_options = aegean_options if aegean_options else AegeanOptions()
    if update_aegean_options:
        aegean_options = aegean_options.with_options(**update_aegean_options)

    image = image.absolute()
    base_output = str(image.parent / image.stem)
    logger.info(f"Using base output name of: {base_output}")

    aegean_names = create_aegean_names(base_output=base_output)
    logger.debug(f"{aegean_names=}")

    bane_command_str = _get_bane_command(image=image, bane_options=bane_options)

    bind_dir = [image.absolute().parent]
    run_singularity_command(
        image=aegean_container,
        command=bane_command_str,
        stream_callback_func=_bane_output_callback,
        bind_dirs=bind_dir,
    )

    aegean_command = _get_aegean_command(
        image=image, base_output=base_output, aegean_options=aegean_options
    )
    run_singularity_command(
        image=aegean_container, command=aegean_command, bind_dirs=bind_dir
    )

    # These are the bane outputs
    bkg_image_path = aegean_names.bkg_image
    rms_image_path = aegean_names.rms_image

    image_header = fits.getheader(image)
    image_beam = (
        image_header["BMAJ"],
        image_header["BMIN"],
        image_header["BPA"],
    )

    aegean_outputs = AegeanOutputs(
        bkg=bkg_image_path,
        rms=rms_image_path,
        comp=aegean_names.comp_cat,
        beam_shape=image_beam,
        image=image,
    )

    logger.info(f"Aegeam finished running. {aegean_outputs=}")

    return aegean_outputs


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(description=__doc__)
    subparsers = parser.add_subparsers(dest="mode")

    bane_parser = subparsers.add_parser(
        name="find", help="Run BANE with default options. "
    )

    bane_parser.add_argument(
        "image", type=Path, help="The image that BANE will process"
    )
    bane_parser.add_argument(
        "container", type=Path, help="Path to container with AegeanTools"
    )
    bane_parser = add_options_to_parser(parser=bane_parser, options_class=BANEOptions)
    bane_parser = add_options_to_parser(parser=bane_parser, options_class=AegeanOptions)

    return parser


def cli() -> None:
    parser = get_parser()

    args = parser.parse_args()

    if args.mode == "find":
        bane_options = create_options_from_parser(
            parser_namespace=args, options_class=BANEOptions
        )
        aegean_options = create_options_from_parser(
            parser_namespace=args, options_class=AegeanOptions
        )
        run_bane_and_aegean(
            image=args.image,
            aegean_container=args.container,
            bane_options=bane_options,
            aegean_options=aegean_options,
        )
    else:
        logger.info(f"Mode '{args.mode}' is not known.")
        parser.print_help()


if __name__ == "__main__":
    cli()



"""Pipeline to calibrate a ASKAP style bandpass observation.

At the moment this is expected to be performed against observations taken
towards PKS1934-638. At the time of writing a typical ASKAP bandpass
observation will cycle each beam so that it is centred on this source.
This means that practically there are 36 separate fields at slightly
different field centres. The bandpass calibration process will first have
to split the correct field out before actually calibration.
"""

from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path
from typing import Collection

from prefect import flow, task, unmapped

from flint.bandpass import extract_correct_bandpass_pointing
from flint.calibrate.aocalibrate import (
    ApplySolutions,
    CalibrateCommand,
    create_apply_solutions_cmd,
    create_calibrate_cmd,
    flag_aosolutions,
    select_aosolution_for_ms,
)
from flint.flagging import flag_ms_aoflagger
from flint.logging import logger
from flint.ms import preprocess_askap_ms, split_by_field
from flint.naming import get_sbid_from_path
from flint.options import (
    MS,
    BandpassOptions,
    add_options_to_parser,
    create_options_from_parser,
)
from flint.prefect.clusters import get_dask_runner
from flint.prefect.common.utils import upload_image_as_artifact
from flint.sky_model import get_1934_model

# These are generic functions that are wrapped. Their inputs are fairly standard
# and do not require any type of unpacking or testing before they are use.
task_extract_correct_bandpass_pointing = task(extract_correct_bandpass_pointing)
task_preprocess_askap_ms = task(preprocess_askap_ms)
task_flag_ms_aoflagger = task(flag_ms_aoflagger)
task_create_calibrate_cmd = task(create_calibrate_cmd)
task_split_by_field = task(split_by_field)
task_select_solution_for_ms = task(select_aosolution_for_ms)
task_create_apply_solutions_cmd = task(create_apply_solutions_cmd)


# The tasks below are ones that require some of the inputs to be cast or transformed
# into something that is known to the actual worker functions.


@task
def task_bandpass_create_apply_solutions_cmd(
    ms: MS,
    calibrate_cmd: CalibrateCommand,
    container: Path,
    output_column: str | None = None,
) -> ApplySolutions:
    """Apply an ao-calibrate style solutions file to an input measurement set.

    Internally the solutions path to apply to the nominaled measurement set is extracted
    from the incoming ``calibrate_cmd``.

    Args:
        ms (MS): The measurement set that will have solutions applied
        calibrate_cmd (CalibrateCommand): The calibrate command and meta-data describing the solutions to apply
        container (Path): Path to singularity container that will apply the solutions
        output_column (Optional[Path], optional): the output column name to create. Defaults to None.

    Returns:
        ApplySolutions: The apply solutions command and meta-data
    """
    return create_apply_solutions_cmd(
        ms=ms,
        solutions_file=calibrate_cmd.solution_path,
        output_column=output_column,
        container=container,
    )


@task
def task_flag_solutions(
    calibrate_cmd: CalibrateCommand,
    **kwargs,
) -> CalibrateCommand:
    """Flag calibration solutions

    Args:
        calibrate_cmd (CalibrateCommand): Calibrate command that contains path to the solution file that will be flagged

    Returns:
        CalibrateCommand: Calibrate command with update meta-data describing the new solutions file
    """
    solution_path = calibrate_cmd.solution_path
    ms_path = calibrate_cmd.ms.path

    plot_dir = ms_path.parent / Path("preflagger")
    if not plot_dir.exists():
        try:
            logger.info(f"Attempting to create {plot_dir}")
            plot_dir.mkdir(parents=True)
        except FileExistsError:
            logger.warning(
                "Creating the directory failed. Likely already exists. Race conditions, me-hearty."
            )

    flagged_solutions = flag_aosolutions(
        solutions_path=solution_path,
        ref_ant=-1,
        flag_cut=3,
        plot_dir=plot_dir,
        **kwargs,
    )

    for image_path in flagged_solutions.plots:
        upload_image_as_artifact(image_path=image_path, description=image_path.name)

    return calibrate_cmd.with_options(
        solution_path=flagged_solutions.path, preflagged=True
    )


def run_bandpass_stage(
    bandpass_mss: Collection[MS],
    output_split_bandpass_path: Path,
    bandpass_options: BandpassOptions,
    model_path: Path,
    source_name_prefix: str = "B1934-638",
    skip_rotation: bool = False,
) -> list[CalibrateCommand]:
    """Executes the bandpass calibration (using ``calibrate``) against a set of
    input measurement sets.

    Args:
        bandpass_mss (Collection[MS]): Set of bandpass measurement sets to calibrate
        output_split_bandpass_path (Path): The location where the extract field centred on the calibration field (typically PKSB19340638)
        bandpass_options (BandpassOptions): Configurables that will specify the bandpass calibbration process
        model_path (Path): Path to the model used to calibrate against
        source_name_prefix (str, optional): Name of the field being calibrated. Defaults to "B1934-638".
        skip_rotation (bool, optional): If ``True`` the rotation of the ASKAP visibility from the antenna frame to the sky-frame will be skipped. Defaults to False.

    Returns:
        List[CalibrateCommand]: Set of calibration commands used
    """
    assert bandpass_options.flag_calibrate_rounds >= 0, (
        f"Currently {bandpass_options.flag_calibrate_rounds=}, needs to be 0 or higher"
    )

    if not output_split_bandpass_path.exists():
        logger.info(f"Creating {output_split_bandpass_path!s}")
        output_split_bandpass_path.mkdir(parents=True)

    calibrate_cmds: list[CalibrateCommand] = []

    extract_bandpass_mss = task_extract_correct_bandpass_pointing.map(
        ms=bandpass_mss,
        source_name_prefix=source_name_prefix,
        ms_out_dir=output_split_bandpass_path,
    )
    preprocess_bandpass_mss = task_preprocess_askap_ms.map(
        ms=extract_bandpass_mss, skip_rotation=skip_rotation
    )
    flag_bandpass_mss = task_flag_ms_aoflagger.map(
        ms=preprocess_bandpass_mss,
        container=bandpass_options.flagger_container,
    )
    calibrate_cmds = task_create_calibrate_cmd.map(
        ms=flag_bandpass_mss,
        calibrate_model=model_path,
        container=bandpass_options.calibrate_container,
        update_calibrate_options=unmapped(dict(minuv=bandpass_options.minuv)),
    )

    for i in range(bandpass_options.flag_calibrate_rounds):
        # Apply and then recalibrate
        apply_cmds = task_bandpass_create_apply_solutions_cmd.map(
            ms=calibrate_cmds,
            calibrate_cmd=calibrate_cmds,
            output_column="CORRECTED_DATA",
            container=bandpass_options.calibrate_container,
        )
        flag_bandpass_mss = task_flag_ms_aoflagger.map(
            ms=apply_cmds, container=bandpass_options.flagger_container
        )
        calibrate_cmds = task_create_calibrate_cmd.map(
            ms=flag_bandpass_mss,
            calibrate_model=model_path,
            container=bandpass_options.calibrate_container,
            calibrate_data_column="DATA",
            update_calibrate_options=unmapped(dict(minuv=bandpass_options.minuv)),
        )
    flag_calibrate_cmds = task_flag_solutions.map(
        calibrate_cmd=calibrate_cmds,
        smooth_solutions=bandpass_options.smooth_solutions,
        smooth_window_size=bandpass_options.smooth_window_size,
        smooth_polynomial_order=bandpass_options.smooth_polynomial_order,
        mean_ant_tolerance=bandpass_options.preflagger_ant_mean_tolerance,
        mesh_ant_flags=bandpass_options.preflagger_mesh_ant_flags,
        max_gain_amplitude=bandpass_options.preflagger_jones_max_amplitude,
    )

    return flag_calibrate_cmds


@flow
def calibrate_bandpass_flow(
    bandpass_path: Path,
    split_path: Path,
    bandpass_options: BandpassOptions,
) -> Path:
    """Create and run the prefect flow to calibrate a set of bandpass measurement sets.

    The measurement sets that will be calibreated are expected to:
    - be following the raw name format convention
    - reside in a directory whose name is the SBID of the observation

    The current bandpass procedure currently relies on the Andre Offringa's ``calibrate`` tool,
    with slight modification from Emil Lenc. The well known source PKS1934-638 is the only source
    supported for bandpass calibration, and its model is packaged inside ``flint``. This model
    is in the AO-style format.

    Each measurement set will correspond to a solutions file once the ``calibrate`` tool has been
    executed successfully. These should be kept together -- there is not enough meta-data in the
    single solutions file to denote the frequency / channels / beam number described in the measurement set.

    Args:
        bandpass_path (Path): Location to the folder containing the raw ASKAP bandpass measurement sets that will be calibrated
        split_path (Path): Location that will contain a folder, named after the SBID of the observation, that will contain the output bandpass measurement sets, solutions and plots
        bandpass_options (BandpassOptions): Options that specify configurables of the bandpass processing.

    Returns:
        Path: Directory that contains the extracted measurement sets and the ao-style gain solutions files.
    """
    assert bandpass_path.exists() and bandpass_path.is_dir(), (
        f"{bandpass_path!s} does not exist or is not a folder. "
    )
    bandpass_mss = list([MS.cast(ms_path) for ms_path in bandpass_path.glob("*.ms")])

    assert len(bandpass_mss) == bandpass_options.expected_ms, (
        f"Expected to find {bandpass_options.expected_ms} in {bandpass_path!s}, found {len(bandpass_mss)}."
    )

    logger.info(
        f"Found the following bandpass measurement set: {[bp.path for bp in bandpass_mss]}."
    )

    bandpass_folder_name = bandpass_path.name
    output_split_bandpass_path = (
        Path(split_path / bandpass_folder_name).absolute().resolve()
    )
    logger.info(
        f"Will write extracted bandpass MSs to: {output_split_bandpass_path!s}."
    )

    # This is the model that we will calibrate the bandpass against.
    # At the time of writing 1934-638 is the only model that is supported,
    # not only by this pirate ship, but also the ASKAP telescope itself.
    model_path: Path = get_1934_model(mode="calibrate")
    source_name_prefix: str = "B1934-638"

    run_bandpass_stage(
        bandpass_mss=bandpass_mss,
        output_split_bandpass_path=output_split_bandpass_path,
        bandpass_options=bandpass_options,
        model_path=model_path,
        source_name_prefix=source_name_prefix,
        skip_rotation=True,
    )

    return output_split_bandpass_path


def setup_run_bandpass_flow(
    bandpass_path: Path,
    split_path: Path,
    cluster_config: Path,
    bandpass_options: BandpassOptions,
) -> Path:
    """Create and run the prefect flow to calibrate a set of bandpass measurement sets.

    The measurement sets that will be calibreated are expected to:
    - be following the raw name format convention
    - reside in a directory whose name is the SBID of the observation

    The current bandpass procedure currently relies on the Andre Offringa's ``calibrate`` tool,
    with slight modification from Emil Lenc. The well known source PKS1934-638 is the only source
    supported for bandpass calibration, and its model is packaged inside ``flint``. This model
    is in the AO-style format.

    Each measurement set will correspond to a solutions file once the ``calibrate`` tool has been
    executed successfully. These should be kept together -- there is not enough meta-data in the
    single solutions file to denote the frequency / channels / beam number described in the measurement set.

    Args:
        bandpass_path (Path): Location to the folder containing the raw ASKAP bandpass measurement sets that will be calibrated
        split_path (Path): Location that will contain a folder, named after the SBID of the observation, that will contain the output bandpass measurement sets, solutions and plots
        cluster_config (Path): Path to a yaml file that is used to configure a prefect dask task runner.
        bandpass_options (BandpassOptions): Options that specify configurables of the bandpass processing.

    Returns:
        Path: Directory that contains the extracted measurement sets and the ao-style gain solutions files.
    """

    dask_task_runner = get_dask_runner(cluster=cluster_config)

    bandpass_sbid = get_sbid_from_path(path=bandpass_path)

    calibrate_bandpass_flow.with_options(
        name=f"Flint Bandpass Pipeline -- {bandpass_sbid}", task_runner=dask_task_runner
    )(
        bandpass_path=bandpass_path,
        split_path=split_path,
        bandpass_options=bandpass_options,
    )

    return bandpass_path


def get_parser() -> ArgumentParser:
    """Create an argument parser for the bandpass prefect workflow

    Returns:
        ArgumentParser: CLI argument parser
    """
    parser = ArgumentParser(
        description="Perform bandpass calibration against an ASKAP SBID. "
    )

    parser.add_argument(
        "bandpass_path",
        type=Path,
        help="Path to the directory containing the uncalibrated bandpass measurement sets. ",
    )
    parser.add_argument(
        "--split-path",
        type=Path,
        default=Path("."),
        help="Location to write the field-split MSs. Will attempt to create a directory using the SBID of the bandpass observation. ",
    )

    parser.add_argument(
        "--cluster-config",
        type=str,
        default="petrichor",
        help="Path to a cluster configuration file, or a known cluster name. ",
    )

    parser = add_options_to_parser(parser=parser, options_class=BandpassOptions)

    return parser


def cli() -> None:
    import logging

    logger.setLevel(logging.INFO)

    parser = get_parser()

    args = parser.parse_args()

    bandpass_options = create_options_from_parser(
        parser_namespace=args, options_class=BandpassOptions
    )

    setup_run_bandpass_flow(
        bandpass_path=args.bandpass_path,
        split_path=args.split_path,
        cluster_config=args.cluster_config,
        bandpass_options=bandpass_options,
    )


if __name__ == "__main__":
    cli()



"""This is a workflow to subtract a continuum model and image the channel-wise data

Unlike the continuum imaging and self-calibnration pipeline this flow currently
expects that all measurement sets are in the flint format, which means other than
the naming scheme that they have been been preprocessed to place them in the IAU
frame and have had their fields table updated. That is to say that they have
already been preprocessed and fixed.
"""

from __future__ import annotations

from pathlib import Path
from time import sleep

import numpy as np
from configargparse import ArgumentParser
from fitscube.combine_fits import combine_fits
from prefect import flow, task, unmapped

from flint.coadd.linmos import LinmosResult
from flint.configuration import get_options_from_strategy, load_and_copy_strategy
from flint.exceptions import FrequencyMismatchError
from flint.logging import logger
from flint.ms import (
    MS,
    consistent_ms_frequencies,
    find_mss,
    get_freqs_from_ms,
)
from flint.naming import get_sbid_from_path
from flint.options import (
    AddModelSubtractFieldOptions,
    SubtractFieldOptions,
    add_options_to_parser,
    create_options_from_parser,
)
from flint.predict.crystalball import CrystalBallOptions
from flint.prefect.clusters import get_dask_runner
from flint.prefect.common.imaging import (
    convolve_then_linmos,
    task_get_common_beam_from_results,
    task_wsclean_imager,
)
from flint.prefect.common.ms import task_subtract_model_from_ms
from flint.prefect.common.predict import task_addmodel_to_ms, task_crystalball_to_ms


def _check_and_verify_options(
    subtract_field_options: SubtractFieldOptions | None = None,
    addmodel_subtract_field_options: AddModelSubtractFieldOptions | None = None,
    crystalball_subtract_field_options: CrystalBallOptions | None = None,
) -> None:
    """Verify that the options supplied to run the subtract field options make sense"""
    if subtract_field_options:
        assert (
            subtract_field_options.wsclean_container.exists()
            and subtract_field_options.wsclean_container.is_file()
        ), (
            f"{subtract_field_options.wsclean_container=} does not exist or is not a file"
        )
        assert (
            subtract_field_options.yandasoft_container.exists()
            and subtract_field_options.yandasoft_container.is_file()
        ), (
            f"{subtract_field_options.yandasoft_container=} does not exist or is not a file"
        )

    if subtract_field_options and subtract_field_options.use_addmodel:
        assert addmodel_subtract_field_options, (
            "Addmodel usage invoked by no AddModelOptions supplied"
        )
        assert addmodel_subtract_field_options.calibrate_container is not None, (
            "Calibrate container path is needede for addmodel"
        )
        assert (
            addmodel_subtract_field_options.calibrate_container.exists()
            and addmodel_subtract_field_options.calibrate_container.is_file()
        ), (
            f"Calibrate container {addmodel_subtract_field_options.calibrate_container} is not a file"
        )
        assert addmodel_subtract_field_options.addmodel_cluster_config is not None, (
            f"{addmodel_subtract_field_options.addmodel_cluster_config=}, which should not happen"
        )

    if subtract_field_options:
        assert (
            sum(
                [
                    subtract_field_options.use_addmodel,
                    subtract_field_options.use_crystalball,
                ]
            )
            <= 1
        ), "Both addmodel and crystallball selected. This surely can not be right. "


def find_mss_to_image(
    mss_parent_path: Path,
    expected_ms_count: int | None = None,
    data_column: str = "CORRECTED_DATA",
    model_column: str = "MODEL_DATA",
) -> tuple[MS, ...]:
    """Search for MSs to image. See ``flint.ms.find_mss`` for further details.

    Args:
        mss_parent_path (Path): Path to search for MSs in
        expected_ms_count (Optional[int], optional): Expected number of measurement sets to find. Defaults to None.
        data_column (str, optional): The nominated data column that should eb set. Defaults to "CORRECTED_DATA".
        model_column (str, optional): The nominated model data column that should be set. Defaults to "MODEL_DATA".

    Returns:
        Tuple[MS, ...]: Collect of MSs
    """
    science_mss = find_mss(
        mss_parent_path=mss_parent_path,
        expected_ms_count=expected_ms_count,
        data_column=data_column,
        model_column=model_column,
    )
    logger.info(f"Found {science_mss=}")
    return science_mss


def find_and_setup_mss(
    science_path_or_mss: Path | tuple[MS, ...],
    expected_ms_count: int,
    data_column: str,
) -> tuple[MS, ...]:
    """Search for MSs in a directory and, if necessary, perform checks around
    their consistency. If the input data appear to be collection of MSs already
    assume they have already been set and checked for consistency.

    Args:
        science_path_or_mss (Union[Path, List[MS, ...]]): Path to search or existing MSs
        expected_ms_count (int): Expected number of MSs to find
        data_column (str): The data column to nominate if creating MSs after searching

    Raises:
        FrequencyMismatchError: Raised when frequency information is not consistent

    Returns:
        Tuple[MS, ...]: Collection of MSs
    """

    if isinstance(science_path_or_mss, (list, tuple)):
        logger.info("Already loaded MSs")
        return tuple(sms for sms in science_path_or_mss)

    # Find the MSs
    # - optionally untar?
    science_mss = find_mss_to_image(
        mss_parent_path=science_path_or_mss,
        expected_ms_count=expected_ms_count,
        data_column=data_column,
    )

    # 2 - ensure matchfing frequencies over channels
    consistent_frequencies_across_mss = consistent_ms_frequencies(mss=science_mss)
    if not consistent_frequencies_across_mss:
        logger.critical("Mismatch in frequencies among provided MSs")
        raise FrequencyMismatchError("There is a mismatch in frequencies")

    return science_mss


@task
def task_combine_all_linmos_images(
    linmos_commands: list[LinmosResult],
    remove_original_images: bool = False,
    combine_weights: bool = False,
) -> Path:
    output_cube_path = Path("test.fits")

    if combine_weights:
        logger.info("Combining weight fits files")
        images_to_combine = [
            linmos_command.weight_fits for linmos_command in linmos_commands
        ]
        output_suffix = "weight"
    else:
        logger.info("Combining image fits files")
        images_to_combine = [
            linmos_command.image_fits for linmos_command in linmos_commands
        ]
        output_suffix = "linmos"

    logger.info(f"Combining {len(images_to_combine)} FITS files together")

    from flint.naming import create_image_cube_name, create_name_from_common_fields

    assert len(images_to_combine) > 0, "No images to combine"

    base_cube_path = create_name_from_common_fields(in_paths=tuple(images_to_combine))
    output_cube_path = create_image_cube_name(
        image_prefix=base_cube_path, mode="contsub", suffix=output_suffix
    )

    _ = combine_fits(
        file_list=images_to_combine,
        out_cube=output_cube_path,
        max_workers=4,
    )

    if remove_original_images:
        logger.info(f"Removing original {len(images_to_combine)} images")
        for image in images_to_combine:
            logger.info(f"Removing {image=}")
            assert isinstance(image, Path) and image.exists(), (
                f"{image=} does not exist, but it should"
            )
            image.unlink()
    return Path(output_cube_path)


@flow
def flow_addmodel_to_mss(
    science_path_or_mss: Path | tuple[MS, ...],
    addmodel_subtract_field_options: AddModelSubtractFieldOptions,
    expected_ms: int,
    data_column: str,
) -> tuple[MS, ...]:
    """Separate flow to perform the potentially expensive model prediction
    into MSs"""
    _check_and_verify_options(
        addmodel_subtract_field_options=addmodel_subtract_field_options
    )

    # Get the MSs that will have their model added to
    science_mss = find_and_setup_mss(
        science_path_or_mss=science_path_or_mss,
        expected_ms_count=expected_ms,
        data_column=data_column,
    )
    science_mss = task_addmodel_to_ms.map(
        ms=science_mss,
        addmodel_subtract_options=unmapped(addmodel_subtract_field_options),
    )

    return science_mss


@flow
def flow_subtract_cube(
    science_path: Path,
    subtract_field_options: SubtractFieldOptions,
    addmodel_subtract_field_options: AddModelSubtractFieldOptions,
    crystalball_subtract_field_options: CrystalBallOptions,
) -> None:
    strategy = load_and_copy_strategy(
        output_split_science_path=science_path,
        imaging_strategy=subtract_field_options.imaging_strategy,
    )
    _check_and_verify_options(
        subtract_field_options=subtract_field_options,
        addmodel_subtract_field_options=addmodel_subtract_field_options,
        crystalball_subtract_field_options=crystalball_subtract_field_options,
    )

    # Find the MSs
    # - optionally untar?
    science_mss = find_and_setup_mss(
        science_path_or_mss=science_path,
        expected_ms_count=subtract_field_options.expected_ms,
        data_column=subtract_field_options.data_column,
    )

    # 2.5 - Continuum subtract if requested

    freqs_mhz = get_freqs_from_ms(ms=science_mss[0]) / 1e6
    logger.info(
        f"Considering {len(freqs_mhz)} frequencies from {len(science_mss)} channels, minimum {np.min(freqs_mhz)}-{np.max(freqs_mhz)}"
    )
    if len(freqs_mhz) > 20 and subtract_field_options.stagger_delay_seconds is None:
        logger.critical(
            f"{len(freqs_mhz)} channels and no stagger delay set! Consider setting a stagger delay"
        )

    if subtract_field_options.use_addmodel:
        assert addmodel_subtract_field_options.addmodel_cluster_config is not None, (
            f"{addmodel_subtract_field_options.addmodel_cluster_config=}, which should not happen"
        )
        addmodel_dask_runner = get_dask_runner(
            cluster=addmodel_subtract_field_options.addmodel_cluster_config
        )
        science_mss = flow_addmodel_to_mss.with_options(
            task_runner=addmodel_dask_runner, name="Predict -- Addmodel"
        )(
            science_path_or_mss=science_mss,
            addmodel_subtract_field_options=addmodel_subtract_field_options,
            expected_ms=subtract_field_options.expected_ms,
            data_column=subtract_field_options.data_column,
        )

    if subtract_field_options.use_crystalball:
        logger.info("Attempting to peer into the crystalball, me'hearty")
        science_mss = task_crystalball_to_ms.map(
            ms=science_mss,
            crystalball_options=unmapped(crystalball_subtract_field_options),
        )

    if subtract_field_options.attempt_subtract:
        science_mss = task_subtract_model_from_ms.map(
            ms=science_mss,
            data_column=subtract_field_options.subtract_data_column,
            update_tracked_column=True,
        )

    channel_parset_list = []
    for channel, freq_mhz in enumerate(freqs_mhz):
        logger.info(f"Imaging {channel=} {freq_mhz=}")
        channel_range = (channel, channel + 1)
        channel_wsclean_cmds = task_wsclean_imager.with_options(retries=2).map(
            in_ms=science_mss,
            wsclean_container=subtract_field_options.wsclean_container,
            channel_range=unmapped(channel_range),
            update_wsclean_options=unmapped(
                get_options_from_strategy(
                    strategy=strategy,
                    mode="wsclean",
                    operation="subtractcube",
                )
            ),
        )
        channel_beam_shape = task_get_common_beam_from_results.submit(
            wsclean_results=channel_wsclean_cmds,
            cutoff=subtract_field_options.beam_cutoff,
            filter_str="image.",
        )
        channel_parset = convolve_then_linmos(
            wsclean_results=channel_wsclean_cmds,
            beam_shape=channel_beam_shape,
            linmos_suffix_str=None,
            field_options=subtract_field_options,
            convol_mode="image",
            convol_filter="image.",
            convol_suffix_str="optimal.conv",
            trim_linmos_fits=False,  # This is necessary to ensure all images have same pixel-coordinates
            remove_original_images=True,
            cleanup_linmos=True,
        )
        channel_parset_list.append(channel_parset)

        if subtract_field_options.stagger_delay_seconds:
            sleep(subtract_field_options.stagger_delay_seconds)

    # 4 - cube concatenated each linmos field together to single file
    task_combine_all_linmos_images.submit(
        linmos_commands=channel_parset_list, remove_original_images=True
    )
    task_combine_all_linmos_images.submit(
        linmos_commands=channel_parset_list,
        remove_original_images=True,
        combine_weights=True,
    )

    return


def setup_run_subtract_flow(
    science_path: Path,
    subtract_field_options: SubtractFieldOptions,
    addmodel_subtract_field_options: AddModelSubtractFieldOptions,
    crystalball_subtract_field_options: CrystalBallOptions,
    cluster_config: Path,
) -> None:
    logger.info(f"Processing {science_path=}")
    science_sbid = get_sbid_from_path(path=science_path)

    dask_runner = get_dask_runner(cluster=cluster_config)

    flow_subtract_cube.with_options(
        task_runner=dask_runner, name=f"Subtract Cube Pipeline -- {science_sbid}"
    )(
        science_path=science_path,
        subtract_field_options=subtract_field_options,
        addmodel_subtract_field_options=addmodel_subtract_field_options,
        crystalball_subtract_field_options=crystalball_subtract_field_options,
    )


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        "--cli-config", is_config_file=True, help="Path to configuration file"
    )
    parser.add_argument(
        "science_path",
        type=Path,
        help="Path to the directory containing the beam-wise measurement sets",
    )
    parser.add_argument(
        "--cluster-config",
        type=str,
        default="petrichor",
        help="Path to a cluster configuration file, or a known cluster name. ",
    )

    parser = add_options_to_parser(parser=parser, options_class=SubtractFieldOptions)
    parser = add_options_to_parser(
        parser=parser, options_class=AddModelSubtractFieldOptions
    )
    parser = add_options_to_parser(parser=parser, options_class=CrystalBallOptions)

    return parser


def cli() -> None:
    parser = get_parser()

    args = parser.parse_args()

    subtract_field_options = create_options_from_parser(
        parser_namespace=args, options_class=SubtractFieldOptions
    )
    addmodel_subtract_field_options = create_options_from_parser(
        parser_namespace=args, options_class=AddModelSubtractFieldOptions
    )
    crystalball_options = create_options_from_parser(
        parser_namespace=args, options_class=CrystalBallOptions
    )

    if addmodel_subtract_field_options.addmodel_cluster_config is None:
        addmodel_subtract_field_options.with_options(
            addmodel_cluster_config=args.cluster_config
        )

    setup_run_subtract_flow(
        science_path=args.science_path,
        subtract_field_options=subtract_field_options,
        addmodel_subtract_field_options=addmodel_subtract_field_options,
        crystalball_subtract_field_options=crystalball_options,
        cluster_config=args.cluster_config,
    )


if __name__ == "__main__":
    cli()



"""Items in and out the model visibility prediction using the `crystalball` python package."""

from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path
from typing import Any

from crystalball.crystalball import predict
from dask.distributed import Client

from flint.imager.wsclean import get_wsclean_output_source_list_path
from flint.logging import logger
from flint.ms import MS
from flint.options import BaseOptions, add_options_to_parser, create_options_from_parser


class CrystalBallOptions(BaseOptions):
    """Options related to running crystal ball"""

    crystallball_wsclean_pol_mode: list[str] = ["i"]
    """The polarisation of the wsclean model that was generated"""
    row_chunks: int = 0
    "Number of rows of input MS that are processed in a single chunk. If 0 it will be set automatically. Default is 0."
    model_chunks: int = 0
    "Number of sky model components that are processed in a single chunk. If 0 it will be set automatically. Default is 0."


def crystalball_predict(
    ms: MS,
    crystalball_options: CrystalBallOptions,
    wsclean_source_list_path: Path | None = None,
    dask_client: Client | None = None,
    output_column: str = "MODEL_DATA",
    update_crystalball_options: dict[str, Any] | None = None,
) -> MS:
    """A very simply wrapper around the `Crystalball.predict` function. Basic
    checks to ensure that the BB6 style source model path exists, which is the
    format used by the `wsclean -save-source-list` option.

    If no `wsclean_source_list_path` is specified one is guess from the name of the
    input MS.path.

    Args:
        ms (MS): The MS instance whose path is to the measurement set to predict into.
        crystalball_options (CrystalBallOptions): Options that control the `crystalball.predict` call.
        wsclean_source_list_path (Path | None, optional): The path to the file with the model components to predict. If None an attempt is made to find it from the MS. Defaults to None.
        dask_client (Client | None, optional): A specialised Dask distributed task. If None one is created by `crystalball`. Defaults to None.
        output_column (str, optional): The column to predict into. The `MS.model_column` will reflect this. Defaults to "MODEL_DATA".
        update_crystalball_options (dict[str, Any] | None, optional): Update options to the provided crystalball_options. Defaults to None.

    Returns:
        MS: The MS that was predicted into, with the `model_column` set appropriately.
    """
    if update_crystalball_options:
        crystalball_options = crystalball_options.with_options(
            **update_crystalball_options
        )

    if wsclean_source_list_path is None:
        assert len(crystalball_options.crystallball_wsclean_pol_mode) == 1, (
            "Only a single polarisation mode is currently supported."
        )

        pol = crystalball_options.crystallball_wsclean_pol_mode[0]
        logger.info(f"Using {pol=}")
        wsclean_source_list_path = get_wsclean_output_source_list_path(
            name_path=ms.path, pol=pol
        )

    assert isinstance(wsclean_source_list_path, Path), (
        f"{wsclean_source_list_path=}, which appears not to be a Path"
    )
    if not wsclean_source_list_path.exists():
        message = f"{wsclean_source_list_path=} was requested, but does not exist"
        raise FileNotFoundError(message)

    logger.info(f"Adding {wsclean_source_list_path=} to {ms.path=}")
    logger.info(f"Predicting into {output_column=} with Crystalball")

    if dask_client:
        logger.info(f"Using {dask_client=}")

    predict(
        ms=str(ms.path),
        sky_model=str(wsclean_source_list_path),
        output_column=output_column,
        client=dask_client,
        row_chunks=crystalball_options.row_chunks,
        model_chunks=crystalball_options.model_chunks,
    )

    return ms.with_options(model_column="MODEL_DATA")


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="The crystal ball interface to add model visibilities from a wsclean file"
    )

    parser.add_argument(
        "ms",
        type=Path,
        help="Path to the measurement set that will have nidek vusuvukutues added",
    )
    parser.add_argument(
        "--model-column",
        type=str,
        help="The name of the model data column",
        default="MODEL_DATA",
    )
    parser.add_argument(
        "--wsclean-source-list-path",
        type=Path,
        default=None,
        help="Path to the model to load. If None one is guessed from the MS name.",
    )

    parser = add_options_to_parser(parser=parser, options_class=CrystalBallOptions)

    return parser


def cli() -> None:
    parser = get_parser()

    args = parser.parse_args()

    crystalball_options = create_options_from_parser(
        parser_namespace=args, options_class=CrystalBallOptions
    )
    ms = MS(path=args.ms, model_column=args.model_column)

    crystalball_predict(
        ms=ms,
        crystalball_options=crystalball_options,
        wsclean_source_list_path=args.wsclean_source_list_path,
    )


if __name__ == "__main__":
    cli()



from __future__ import annotations


class FlintException(Exception):
    """Base exception for Flint"""

    pass


class UnknownContainerException(FlintException):
    """An unknown container was requested"""

    pass


class NamingException(FlintException):
    """Raised when there are issues with naming"""

    pass


class NotSupportedError(FlintException):
    """Raised when something is not supported"""

    pass


class AttemptRerunException(FlintException):
    """Intended to be used in stream call back functions to
    capture strange errors and signify that they should be
    rerun. For instance, strange BANE deadlocking errors."""


class TimeLimitException(FlintException):
    """A function has taken too long to execute"""

    pass


class MSError(FlintException):
    """An error for MS related things"""

    pass


class FrequencyMismatchError(FlintException):
    """Raised when there are differences in frequencies"""


class PhaseOutlierFitError(FlintException):
    """Raised when the phase outlier fit routine fails."""

    pass


class GainCalError(FlintException):
    """Raised when it appears like the casa gaincal task fails."""

    pass


class CleanDivergenceError(FlintException):
    """Raised if it is detected that cleaning has diverged."""

    pass


class TarArchiveError(FlintException):
    """Raised it the flint tarball is not created successfullty"""



"""Some tests related to using aoccalibrate related things"""

from __future__ import annotations

import shutil
from pathlib import Path

import numpy as np
import pytest

from flint.bptools.smoother import (
    divide_bandpass_by_ref_ant,
    divide_bandpass_by_ref_ant_preserve_phase,
    smooth_bandpass_complex_gains,
    smooth_data,
)
from flint.calibrate.aocalibrate import (
    AOSolutions,
    CalibrateOptions,
    FlaggedAOSolution,
    calibrate_options_to_command,
    flag_aosolutions,
    plot_solutions,
    select_refant,
)
from flint.utils import get_packaged_resource_path


def test_calibrate_options_to_command():
    default_cal = CalibrateOptions(datacolumn="DATA", m=Path("/example/1934.model"))
    ex_ms_path = Path("/example/data.ms")
    ex_solutions_path = Path("/example/sols.calibrate")

    cmd = calibrate_options_to_command(
        calibrate_options=default_cal,
        ms_path=ex_ms_path,
        solutions_path=ex_solutions_path,
    )

    assert (
        cmd
        == "calibrate -datacolumn DATA -m /example/1934.model -i 100 /example/data.ms /example/sols.calibrate"
    )


def test_calibrate_options_to_command2():
    default_cal = CalibrateOptions(
        datacolumn="DATA",
        m=Path("/example/1934.model"),
        i=40,
        p=(Path("amps.plot"), Path("phase.plot")),
    )
    ex_ms_path = Path("/example/data.ms")
    ex_solutions_path = Path("/example/sols.calibrate")

    cmd = calibrate_options_to_command(
        calibrate_options=default_cal,
        ms_path=ex_ms_path,
        solutions_path=ex_solutions_path,
    )

    assert (
        cmd
        == "calibrate -datacolumn DATA -m /example/1934.model -i 40 -p amps.plot phase.plot /example/data.ms /example/sols.calibrate"
    )


def test_calibrate_options_to_command3():
    default_cal = CalibrateOptions(
        datacolumn="DATA",
        m=Path("/example/1934.model"),
        i=40,
        p=(Path("amps.plot"), Path("phase.plot")),
        maxuv=5000,
        minuv=300,
    )
    ex_ms_path = Path("/example/data.ms")
    ex_solutions_path = Path("/example/sols.calibrate")

    cmd = calibrate_options_to_command(
        calibrate_options=default_cal,
        ms_path=ex_ms_path,
        solutions_path=ex_solutions_path,
    )

    assert (
        cmd
        == "calibrate -datacolumn DATA -m /example/1934.model -minuv 300.0 -maxuv 5000.0 -i 40 -p amps.plot phase.plot /example/data.ms /example/sols.calibrate"
    )


@pytest.fixture
def ao_sols(tmpdir):
    ao_sols = Path(
        get_packaged_resource_path(
            package="flint.data.tests", filename="SB39433.B1934-638.beam0.calibrate.bin"
        )
    )

    out_ao_sols = Path(tmpdir) / ao_sols.name

    shutil.copyfile(ao_sols, out_ao_sols)

    return out_ao_sols


@pytest.fixture
def ao_sols_known_bad(tmpdir):
    # The file contains a binary solutions file that failed previously.
    # It was fixed by testing for all nans in the `flint.bptools.smoother.smooth_data`
    # function.
    ao_sols = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB38969.B1934-638.beam35.aocalibrate.bin",
        )
    )

    out_ao_sols = Path(tmpdir) / ao_sols.name

    shutil.copyfile(ao_sols, out_ao_sols)

    return out_ao_sols


def test_known_bad_sols(ao_sols_known_bad):
    flag_aosolutions(solutions_path=ao_sols_known_bad, plot_solutions_throughout=False)


def test_sols_same_with_plots(ao_sols_known_bad):
    # Had a thought at one point the plktting was updating th mutable numpy array before
    # it was written back to file. Wrote the test, and it passed. Test stays
    a = flag_aosolutions(
        solutions_path=ao_sols_known_bad, plot_solutions_throughout=False
    )
    a_loaded = AOSolutions.load(a.path)

    b = flag_aosolutions(
        solutions_path=ao_sols_known_bad, plot_solutions_throughout=True
    )
    b_loaded = AOSolutions.load(b.path)

    assert np.allclose(a_loaded.bandpass, b_loaded.bandpass, equal_nan=True)


def test_flagged_aosols_mesh(ao_sols_known_bad):
    flagged_sols = flag_aosolutions(
        solutions_path=ao_sols_known_bad,
        plot_solutions_throughout=True,
        smooth_solutions=True,
    )
    assert isinstance(flagged_sols, FlaggedAOSolution)
    assert len(flagged_sols.plots) == 9
    assert isinstance(flagged_sols.path, Path)

    flagged_sols = flag_aosolutions(
        solutions_path=ao_sols_known_bad,
        mesh_ant_flags=True,
        plot_solutions_throughout=False,
        smooth_solutions=False,
    )
    assert isinstance(flagged_sols, FlaggedAOSolution)
    assert len(flagged_sols.plots) == 0
    assert isinstance(flagged_sols.path, Path)


def test_flagged_aosols(ao_sols_known_bad):
    flagged_sols = flag_aosolutions(
        solutions_path=ao_sols_known_bad,
        plot_solutions_throughout=True,
        smooth_solutions=True,
    )
    assert isinstance(flagged_sols, FlaggedAOSolution)
    assert len(flagged_sols.plots) == 9
    assert isinstance(flagged_sols.path, Path)

    flagged_sols = flag_aosolutions(
        solutions_path=ao_sols_known_bad,
        plot_solutions_throughout=True,
        smooth_solutions=False,
    )
    assert isinstance(flagged_sols, FlaggedAOSolution)
    assert len(flagged_sols.plots) == 6
    assert isinstance(flagged_sols.path, Path)


def test_load_aosols(ao_sols):
    ao = AOSolutions.load(path=ao_sols)

    assert ao.nant == 36
    assert ao.nchan == 288
    assert ao.npol == 4


def test_aosols_bandpass_plot(ao_sols):
    # This is just a dumb test to make sure the function runs
    plot_solutions(solutions=ao_sols, ref_ant=0)
    plot_solutions(solutions=ao_sols, ref_ant=None)


def test_aosols_all_nans_smooth_data(ao_sols):
    ao = AOSolutions.load(ao_sols)
    ao.bandpass[0, 20, :, :] = np.nan
    assert np.all(~np.isfinite(ao.bandpass[0, 20, :, 0]))

    smoothed = smooth_data(
        data=ao.bandpass[0, 20, :, 0].real, window_size=16, polynomial_order=4
    )
    assert np.all(~np.isfinite(smoothed))


def test_smooth_bandpass_complex_gains_nans(ao_sols):
    ao = AOSolutions.load(ao_sols)
    ao.bandpass[0, 20, :, :] = np.nan
    assert np.all(~np.isfinite(ao.bandpass[0, 20, :, 0]))

    smoothed = smooth_bandpass_complex_gains(
        complex_gains=ao.bandpass[0], window_size=16, polynomial_order=4
    )
    assert np.all(~np.isfinite(smoothed[20, :, 0]))


def test_smooth_bandpass_complex_gains_nans_with_refant(ao_sols):
    ao = AOSolutions.load(ao_sols)
    ao.bandpass[0, 20, :, :] = np.nan
    assert np.all(~np.isfinite(ao.bandpass[0, 20, :, 0]))

    ref = divide_bandpass_by_ref_ant(complex_gains=ao.bandpass[0], ref_ant=0)

    smoothed = smooth_bandpass_complex_gains(
        complex_gains=ref, window_size=16, polynomial_order=4
    )
    assert np.all(~np.isfinite(smoothed[20, :, 0]))


def test_aosols_bandpass_ref_nu_rank_error(ao_sols):
    ao = AOSolutions.load(path=ao_sols)

    # This should raise an assertion error since the data shape is not right
    with pytest.raises(AssertionError) as _:
        divide_bandpass_by_ref_ant(complex_gains=ao.bandpass, ref_ant=0)


def test_aosols_bandpass_ref_nu(ao_sols):
    ao = AOSolutions.load(path=ao_sols)

    complex_gains = divide_bandpass_by_ref_ant(complex_gains=ao.bandpass[0], ref_ant=0)

    expected = np.array(
        [
            0.11008759 - 0.00000000e00j,
            0.11009675 - 4.33444224e-19j,
            0.11017988 - 0.00000000e00j,
            0.10990718 - 0.00000000e00j,
            0.11060902 - 8.66905258e-19j,
        ]
    )
    assert np.allclose(expected, complex_gains[0, :5, 0])


def test_aosols_bandpass_ref_nu_preserve_phase(ao_sols):
    ao = AOSolutions.load(path=ao_sols)

    complex_gains = divide_bandpass_by_ref_ant_preserve_phase(
        complex_gains=ao.bandpass[0], ref_ant=0
    )

    x_angle = np.angle(complex_gains[0, :, 0])
    y_angle = np.angle(complex_gains[0, :, 3])

    assert np.allclose(x_angle[np.isfinite(x_angle)], 0)
    assert np.allclose(y_angle[np.isfinite(y_angle)], 0)

    expected = np.array(
        [
            -0.10846614 - 0.01465966j,
            -0.10776107 - 0.01495074j,
            -0.10728749 - 0.01611982j,
            -0.10742277 - 0.01654671j,
        ]
    )

    print(expected)
    print(complex_gains[1, :4, 0])

    assert np.allclose(expected, complex_gains[1, :4, 0])


def test_ref_ant_selection(ao_sols):
    ao = AOSolutions.load(path=ao_sols)

    ref_ant = select_refant(bandpass=ao.bandpass)

    assert ref_ant == 0


def test_ref_ant_selection_with_assert(ao_sols):
    ao = AOSolutions.load(path=ao_sols)

    # This ref ant selection function expects a rank of 4
    with pytest.raises(AssertionError) as _:
        select_refant(bandpass=ao.bandpass[0])


# TODO: Need to write more tests for the smoothing and other things
